__author__ = "Simon Nilsson"

import math
import os
from copy import deepcopy
from subprocess import PIPE, Popen
from tkinter import *
from tkinter import filedialog
from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib

matplotlib.use('agg')
import cv2
import pandas as pd
from PIL import Image, ImageTk
from tabulate import tabulate

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import simba
from simba.mixins.config_reader import ConfigReader
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimbaButton, SimbaCheckbox, SimBALabel)
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int, check_str, check_that_column_exist,
                                check_valid_boolean, check_valid_dataframe,
                                check_valid_dict)
from simba.utils.enums import ConfigKey, Dtypes, Formats, Options, TagNames
from simba.utils.errors import FrameRangeError, NoDataError, NoFilesFoundError
from simba.utils.lookups import (get_current_time, get_display_resolution,
                                 get_labelling_img_kbd_bindings,
                                 get_labelling_video_kbd_bindings)
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (get_fn_ext, get_video_meta_data,
                                    read_config_entry, read_config_file,
                                    read_df, read_frm_of_video, write_df)

PLAY_VIDEO_SCRIPT_PATH = os.path.join(os.path.dirname(simba.__file__), "labelling/play_annotation_video.py")
PADDING = 5

DISPLAY_RESOLUTION = get_display_resolution()
MAX_FRAME_SIZE = (int(DISPLAY_RESOLUTION[0] * 0.5), int(DISPLAY_RESOLUTION[1] * 0.05))


class LabellingInterface(ConfigReader):
    """
    Launch ``standard`` or ``pseudo``-labelling (annotation) GUI interface in SimBA.

    .. note::
       Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/label_behavior.md>`__.

    .. image:: _static/img/annotator.png
       :width: 500
       :align: center


    :param Union[str, os.PathLike] config_path: path to SimBA project config file in Configparser format
    :param Union[str, os.PathLike] file_path: Path to video that is to be annotated.
    :param Literal["from_scratch", "pseudo"] setting: String representing annotation method. OPTIONS: ``from_scratch`` or ``pseudo``
    :param Optional[Dict[str, float]] threshold_dict: If setting ``pseudo``, threshold_dict dict contains the machine probability thresholds, with the classifier names as keys and the classification probabilities as values, e.g. {'Attack': 0.40, 'Sniffing': 0.7).
    :param bool continuing: If True, continouing previously started annotation session.

    Examples
    ----------
    >>> select_labelling_video(config_path='MyConfigPath', threshold_dict={'Attack': 0.4}, file_path='MyVideoFilePath', setting='pseudo', continuing=False)
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 file_path: Union[str, os.PathLike],
                 threshold_dict: Optional[Dict[str, float]] = None,
                 setting: Literal["from_scratch", "pseudo"] = "from_scratch",
                 continuing: bool = False):

        ConfigReader.__init__(self, config_path=config_path)
        log_event(logger_name=str(self.__class__.__name__), log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        check_file_exist_and_readable(file_path=file_path)
        self.video_meta_data = get_video_meta_data(video_path=file_path)
        check_str(name='setting', value=setting, options=('pseudo', "from_scratch",))
        check_valid_boolean(value=continuing, source=f'{self.__class__.__name__} continuing')
        if threshold_dict is not None:
            check_valid_dict(x=threshold_dict, valid_key_dtypes=(str,), valid_values_dtypes=(float,), min_value=0.0, max_value=1.0)
        self.img_idx, self.threshold_dict = 0, threshold_dict
        self.setting, self.video_path = setting, file_path
        self.video_name = self.video_meta_data['video_name']
        self.features_extracted_file_path = os.path.join(self.features_dir, f"{self.video_name}.{self.file_type}")
        self.targets_inserted_file_path = os.path.join(self.targets_folder, f"{self.video_name}.{self.file_type}")
        self.machine_results_file_path = os.path.join(self.machine_results_dir, f"{self.video_name}.{self.file_type}")
        self.cap = cv2.VideoCapture(self.video_path)
        self.img_kbd_bindings = get_labelling_img_kbd_bindings()
        self.video_kbd_bindings = get_labelling_video_kbd_bindings()
        self.key_presses_lbl = self.get_keyboard_shortcuts_lbl(self.img_kbd_bindings, header="FRAME NAVIGATION SHORTCUTS: ")
        self.video_presses_lbl = self.get_keyboard_shortcuts_lbl(self.video_kbd_bindings, header="VIDEO NAVIGATION SHORTCUTS: ")
        self.max_frm_id = self.video_meta_data["frame_count"] - 1
        self.temp_file_path = os.path.join(os.path.dirname(self.config_path), "subprocess.txt")
        if len(self.clf_names) == 0:
            raise NoDataError(msg='To annotate behaviors, your SimBA project needs at least one defined classifier. Found 0 classifiers defined in SimBA project', source=self.__class__.__name__)
        self.main_window = Toplevel()






        if continuing:
            if not os.path.isfile(self.targets_inserted_file_path):
                raise NoFilesFoundError(msg=f'When continuing annotations, SimBA expects a file at {self.targets_inserted_file_path}. SimBA could not find this file.', source=self.__class__.__name__)
            self.data_df = read_df(self.targets_inserted_file_path, self.file_type)
            self.data_df_features = read_df(self.features_extracted_file_path, self.file_type)
            missing_idx = self.data_df_features.index.difference(self.data_df.index)
            if len(missing_idx) > 0:
                self.data_df_features = self.data_df_features.iloc[self.data_df_features.index.difference(self.data_df.index)]
                self.data_df = pd.concat([self.data_df.astype(int), self.data_df_features], axis=0).sort_index()
            self.main_window.title("SIMBA ANNOTATION INTERFACE (CONTINUING ANNOTATIONS) - {}".format( self.video_name))
            self.img_idx = read_config_entry(self.config, "Last saved frames", self.video_name, data_type="int", default_value=0)
        else:
            if setting == "from_scratch":
                if not os.path.isfile(self.features_extracted_file_path):
                    raise NoFilesFoundError(msg=f'When annotating data, SimBA expects a file at {self.features_extracted_file_path}. SimBA could not find this file. Extract features for video {self.video_name} before annotating data.', source=self.__class__.__name__)
                self.data_df = read_df(self.features_extracted_file_path, self.file_type)
                self.main_window.title("SIMBA ANNOTATION INTERFACE (ANNOTATING FROM SCRATCH) - {}".format(self.video_name))
                for target in self.clf_names:
                    self.data_df[target] = 0
            elif setting == "pseudo":
                if not os.path.isfile(self.machine_results_file_path):
                    raise NoFilesFoundError(msg=f'When doing pseudo-annotations, SimBA expects a file at {self.machine_results_file_path}. SimBA could not find this file.', source=self.__class__.__name__)
                self.data_df = read_df(self.machine_results_file_path, self.file_type)
                check_valid_dataframe(df=self.data_df, source=self.__class__.__name__, required_fields=self.clf_names)
                for target in self.clf_names:
                    self.data_df.loc[self.data_df[f"Probability_{target}"] > self.threshold_dict[target], target] = 1
                    self.data_df.loc[self.data_df[f"Probability_{target}"] <= self.threshold_dict[target], target] = 0
                self.main_window.title("SIMBA ANNOTATION INTERFACE (PSEUDO-LABELLING) - {}".format(self.video_name))

        self.data_df_targets = self.data_df.copy()
        check_valid_dataframe(df=self.data_df_targets, source=f'{self.__class__.__name__} file_path', valid_dtypes=Formats.NUMERIC_DTYPES.value, required_fields=self.clf_names)
        self.data_df_targets = self.data_df_targets[self.clf_names]

        self.img_frm = CreateLabelFrameWithIcon(parent=self.main_window, header='VIDEO FRAME', icon_name='video', relief='solid')
        self.img_lbl = Label(self.img_frm, name='img_lbl')
        self.img_frm.grid(row=0, column=0, sticky=NW, padx=10, pady=10, columnspan=3, rowspan=2)
        self.img_lbl.grid(row=0, column=0, sticky=NW)
        self.set_img(img_lbl=self.img_lbl, video_path=self.video_path, img_idx=self.img_idx)


        self.navigation_frm = CreateLabelFrameWithIcon(parent=self.main_window, header='FRAME NAVIGATION',  icon_name='navigation', relief='solid')
        self.current_frm_eb = Entry_Box(parent=self.navigation_frm, fileDescription='CURRENT FRAME: ', labelwidth=30, validation='numeric', entry_box_width=20, value=self.img_idx, justify='center', label_font=Formats.FONT_HEADER.value, entry_font=Formats.FONT_HEADER.value)
        self.got_to_frm_btn = SimbaButton(parent=self.navigation_frm, txt="GO TO FRAME", img='walk', font=Formats.FONT_HEADER.value, cmd=self.__advance_frame, cmd_kwargs={'frm_id': lambda: int(self.current_frm_eb.entry_get)}, hover_font=Formats.FONT_HEADER.value)
        self.first_frm_btn = SimbaButton(parent=self.navigation_frm, txt="", tooltip_txt='GO TO FIRST FRAME', img='rewind_large', cmd=self.__advance_frame, cmd_kwargs={'frm_id': 0})
        self.reverse_btn = SimbaButton(parent=self.navigation_frm, txt="", img='reverse_large', tooltip_txt="-1 FRAME", cmd=self.__advance_frame, cmd_kwargs={'frm_id': lambda: self.img_idx - 1})
        self.forward_btn = SimbaButton(parent=self.navigation_frm, txt="", img='forward_large', tooltip_txt="+1 FRAME", cmd=self.__advance_frame, cmd_kwargs={'frm_id': lambda: self.img_idx + 1})
        self.last_frm_btn = SimbaButton(parent=self.navigation_frm, txt="", img='fast_forward_large_2', tooltip_txt="GO TO LAST FRAME", cmd=self.__advance_frame, cmd_kwargs={'frm_id': self.max_frm_id})

        self.navigation_frm.grid(row=2, column=0, sticky=NSEW, pady=10, padx=10, columnspan=1)
        self.current_frm_eb.grid(row=0, column=0, sticky=EW, columnspan=4, padx=5, pady=32)
        self.got_to_frm_btn.grid(row=0, column=5, sticky=EW, columnspan=2, padx=5, pady=32)
        self.first_frm_btn.grid(row=1, column=0, sticky=E, padx=5, pady=5)
        self.reverse_btn.grid(row=1, column=1, sticky=E, padx=5, pady=5)
        self.forward_btn.grid(row=1, column=2, sticky=E, padx=5, pady=5)
        self.last_frm_btn.grid(row=1, column=3, sticky=E, padx=5, pady=5)


        self.jump_frm = CreateLabelFrameWithIcon(parent=self.main_window, header='JUMP TO FRAME',  icon_name='jump', relief='solid')
        self.jump_scale = Scale(self.jump_frm, from_=0, to=100, orient=HORIZONTAL, length=200, label="FRAME JUMP SIZE", tickinterval=25)
        self.jump_scale.set(0)
        self.jump_back = SimbaButton(parent=self.jump_frm, txt="", tooltip_txt="JUMP BACKWARDS", img='pointing_left_large', cmd=self.__advance_frame, cmd_kwargs={'frm_id': lambda: self.img_idx - int(self.jump_scale.get())})
        self.jump_forwards = SimbaButton(parent=self.jump_frm, txt="", img='point_right_large', tooltip_txt='JUMP FORWARDS', cmd=self.__advance_frame, cmd_kwargs={'frm_id': lambda: self.img_idx + int(self.jump_scale.get())})

        self.jump_frm.grid(row=2, column=1, sticky=NSEW, pady=10, padx=10, columnspan=2)
        self.jump_scale.grid(row=0, column=0, sticky=W, padx=5, pady=5, columnspan=2)
        self.jump_back.grid(row=1, column=0, sticky=E, padx=5, pady=5, columnspan=1)
        self.jump_forwards.grid(row=1, column=1, sticky=W, padx=5, pady=5, columnspan=1)

        self.range_frm = CreateLabelFrameWithIcon(parent=self.main_window, header='ANNOTATE FRAME RANGE', icon_name='range_large', relief='solid')
        self.range_start_eb = Entry_Box(parent=self.range_frm, fileDescription='START FRAME: ', labelwidth=30, validation='numeric', entry_box_width=20, justify='center', label_font=Formats.FONT_HEADER.value, entry_font=Formats.FONT_HEADER.value)
        self.range_end_eb = Entry_Box(parent=self.range_frm, fileDescription='RANGE END FRAME: ', labelwidth=30, validation='numeric', entry_box_width=20, justify='center', label_font=Formats.FONT_HEADER.value, entry_font=Formats.FONT_HEADER.value)
        self.save_range_btn = SimbaButton(parent=self.range_frm, txt="SAVE RANGE", img='save_small', cmd=self.__save_behavior_in_range, cmd_kwargs={'start_frm': lambda: self.range_start_eb.entry_get, 'end_frm': lambda: self.range_end_eb.entry_get})
        self.range_frm.grid(row=3, column=0, sticky=NSEW, padx=10, pady=10, columnspan=1)
        self.range_start_eb.grid(row=0, column=0, sticky=NW)
        self.range_end_eb.grid(row=1, column=0, sticky=NW)
        self.save_range_btn.grid(row=2, column=0, sticky=NW)

        self.save_frm = CreateLabelFrameWithIcon(parent=self.main_window, header='SAVE ANNOTATIONS', icon_name='save_large', relief='solid')
        self.save_btn = SimbaButton(parent=self.save_frm, txt="ANNOTATIONS SAVE", tooltip_txt="SAVE ANNOTATIONS TO DISK", img='rocket', cmd=self.__advance_frame, cmd_kwargs={'frm_id': self.max_frm_id}, font=Formats.FONT_HEADER.value)
        self.save_frm.grid(row=3, column=1, sticky=NSEW, padx=10, pady=10, columnspan=2)
        self.save_btn.grid(row=0, column=0, sticky=NW, padx=5,  pady=20)


        self.clf_cb_frm = CreateLabelFrameWithIcon(parent=self.main_window, header='CHECK BEHAVIOR', icon_name='forest', relief='solid')
        self.clf_cbs = self.get_clf_cbs(parent=self.clf_cb_frm, clfs=self.clf_names)
        self.clf_cb_frm.grid(row=0, column=4, sticky=NSEW, padx=10, pady=10)

        self.video_player_frm = CreateLabelFrameWithIcon(parent=self.main_window, header='PLAY VIDEO', icon_name='tv_large', relief='solid')
        self.play_video_btn = SimbaButton(parent=self.video_player_frm, txt="OPEN VIDEO", img='video_file_large', cmd=self.__play_video)
        self.update_img_from_video = SimbaButton(parent=self.video_player_frm, txt="SHOW CURRENT VIDEO FRAME", img='display_large', cmd=self.__update_frame_from_video)
        self.video_player_frm.grid(row=1, column=4, sticky=NSEW, padx=10, pady=10)
        self.play_video_btn.grid(row=0, column=0, sticky=NSEW, padx=3, pady=3)
        self.update_img_from_video.grid(row=0, column=1, sticky=NSEW,  padx=3, pady=3)

        SimBALabel(parent=self.video_player_frm, txt=self.video_presses_lbl, font=Formats.FONT_HEADER.value).grid(row=2, column=0, sticky=NSEW, padx=5, columnspan=2)
        SimBALabel(parent=self.video_player_frm, txt=self.key_presses_lbl, font=Formats.FONT_HEADER.value, justify='center').grid(row=3, column=0, sticky=NSEW, padx=5, columnspan=2)
        self.__bind_frm_shortcut_keys()

        self.main_window.mainloop()

    def __bind_frm_shortcut_keys(self):
        self.main_window.bind(self.img_kbd_bindings['save']['kbd'], lambda x: self.__save_results())
        self.main_window.bind(self.img_kbd_bindings['frame+1_keep_choices']['kbd'], lambda x: self.__advance_frame(frm_id=self.img_idx+1, save_and_keep_checks=True))
        self.main_window.bind(self.img_kbd_bindings['print_annotation_statistic']['kbd'], lambda x: self.print_annotation_statistics(video_meta_data=self.video_meta_data, clf_names=self.clf_names, data_df=self.data_df_targets))
        self.main_window.bind(self.img_kbd_bindings['frame+1']['kbd'], lambda x: self.__advance_frame(frm_id=self.img_idx+1))
        self.main_window.bind(self.img_kbd_bindings['frame-1']['kbd'], lambda x: self.__advance_frame(frm_id=self.img_idx-1))
        self.main_window.bind(self.img_kbd_bindings['last_frame']['kbd'], lambda x: self.__advance_frame(frm_id=self.max_frm_id))
        self.main_window.bind(self.img_kbd_bindings['first_frame']['kbd'], lambda x: self.__advance_frame(0))


    def __play_video(self):
        p = Popen(f"python {PLAY_VIDEO_SCRIPT_PATH}",stdin=PIPE, stdout=PIPE, shell=True)
        p.stdin.write(bytes(self.video_path, "utf-8")); p.stdin.close()
        with open(self.temp_file_path, "w") as text_file: text_file.write(str(p.pid))

    def __update_frame_from_video(self):
        f = open(os.path.join(os.path.dirname(self.config_path), "labelling_info.txt"), "r+")
        os.fsync(f.fileno())
        vid_frame_no = int(f.readline())
        self.__advance_frame(frm_id=vid_frame_no)
        f.close()


    def get_keyboard_shortcuts_lbl(self,
                                   kbd_bindings: Dict[str, Dict[str, str]],
                                   header: str = "FRAME NAVIGATION SHORTCUTS: "):

        labels = [v['label'] for v in kbd_bindings.values()]
        num_items = len(labels)
        mid = math.ceil(num_items / 2)
        col1 = labels[:mid]
        col2 = labels[mid:]
        if len(col2) < len(col1):
            col2 += [''] * (len(col1) - len(col2))
        key_presses_lbl = f"\n\n {header}\n"
        for left, right in zip(col1, col2):
            key_presses_lbl += f"{left:<40} {right}\n"

        return key_presses_lbl



    def set_img(self,
                img_lbl: Label,
                video_path: Union[str, os.PathLike],
                img_idx: int,
                display_size: Optional[Tuple[int, int]] = None):

        self.img = read_frm_of_video(video_path=video_path, frame_index=img_idx, size=display_size)
        img_rgb = cv2.cvtColor(self.img, cv2.COLOR_BGR2RGB)
        self.pil_image = Image.fromarray(img_rgb)
        self.tk_image = ImageTk.PhotoImage(self.pil_image)
        img_lbl.configure(image=self.tk_image)
        img_lbl.image = self.tk_image

    def get_clf_cbs(self,
                    parent: Toplevel,
                    clfs: List[str]) -> Dict[str, BooleanVar]:

        clf_cb_vars = {}
        for clf_cnt, clf_name in enumerate(clfs):
            cb, clf_cb_vars[clf_name] = SimbaCheckbox(parent=parent, txt=clf_name, font=Formats.FONT_HEADER.value)
            cb.grid(row=clf_cnt, column=0, sticky=NW, padx=5, pady=5)
        return clf_cb_vars


    def print_annotation_statistics(self,
                                    video_meta_data: Dict[str, Any],
                                    clf_names: List[str],
                                    data_df: pd.DataFrame):

        table_view = [["Video name", video_meta_data['video_name']],
                      ["Video frames", video_meta_data["frame_count"]],
                      ['Current time:', get_current_time()]]
        for clf in clf_names:
            present = len(data_df[data_df[clf] == 1])
            absent = len(data_df[data_df[clf] == 0])
            table_view.append([clf + " present labels", present])
            table_view.append([clf + " absent labels", absent])
            table_view.append([clf + " % present", present / video_meta_data["frame_count"]])
            table_view.append([clf + " % absent", absent / video_meta_data["frame_count"]])
        headers = ["VARIABLE", "VALUE"]
        print(tabulate(table_view, headers, tablefmt="github"))

    def __advance_frame(self,
                        frm_id: int,
                        save_and_keep_checks: bool = False):

        if frm_id > self.max_frm_id:
            raise FrameRangeError(msg=f"TRYING TO SHOW FRAME {frm_id} WHICH CANNOT BE SHOWN - THE VIDEO {self.video_name} HAS {self.video_meta_data['frame_count']} FRAMES")
        elif frm_id < 0:
            raise FrameRangeError(msg=f"TRYING TO SHOW FRAME {frm_id} WHICH CANNOT BE SHOWN. FRAME INDEX HAS TO BE 0 OR ABOVE")
        else:
            prior_frm_idx = int(self.current_frm_eb.entry_get)
            self.__create_print_statements(frame=prior_frm_idx)
            self.current_frm_eb.entry_set(val=frm_id)
            if not save_and_keep_checks:
                for target in self.clf_names:
                    self.clf_cbs[target].set(bool(self.data_df_targets[target].loc[frm_id]))
            else:
                for target in self.clf_names:
                    self.save_behavior_in_frm(clf_name=target, frame_idx=prior_frm_idx)
                    self.clf_cbs[target].set(bool(self.data_df_targets[target].loc[prior_frm_idx]))

            self.img_idx = deepcopy(frm_id)
            self.set_img(img_lbl=self.img_lbl, video_path=self.video_path, img_idx=frm_id)

    def __save_behavior_in_range(self,
                                 start_frm: int,
                                 end_frm: int):

        check_int(name="RANGE START FRAME", value=start_frm, min_value=0, max_value=self.max_frm_id)
        check_int(name="RANGE END FRAME", value=end_frm, min_value=0, max_value=self.max_frm_id)
        start_frm, end_frm = int(start_frm), int(end_frm)
        if start_frm >= end_frm:
            raise FrameRangeError(msg=f'The RANGE START FRAME ({start_frm}) has to be smaller than the RANGE END FRAME ({end_frm})', source=self.__class__.__name__)
        for frm_no in range(int(start_frm), int(end_frm) + 1):
            for target in self.clf_names:
                self.data_df_targets[target].loc[frm_no] = self.clf_cbs[target].get()
        self.set_img(img_lbl=self.img_lbl, video_path=self.video_path, img_idx=end_frm)
        self.current_frm_eb.entry_set(end_frm)
        self.img_idx = deepcopy(end_frm)
        self.__create_print_statements(frame=(start_frm, end_frm))

    def save_behavior_in_frm(self,
                             frame_idx: int,
                             clf_name: str):

        self.data_df_targets[clf_name].loc[frame_idx] = int(self.clf_cbs[clf_name].get())

    def __save_results(self):
        self.features_df = read_df(self.features_extracted_file_path, self.file_type)
        self.save_df = pd.concat([self.features_df, self.data_df_targets], axis=1)
        try:
            write_df(self.save_df, self.file_type, self.targets_inserted_file_path)
        except Exception as e:
            print(e, f"SIMBA ERROR: File for video {get_fn_ext(self.features_extracted_file_path)[1]} could not be saved.")
            raise FileExistsError
        stdout_success(msg=f"SAVED: Annotation file for video {self.video_name} saved within the {self.targets_folder} directory.", source=self.__class__.__name__)
        if not self.config.has_section("Last saved frames"):
            self.config.add_section("Last saved frames")
        self.config.set("Last saved frames", str(self.video_name), str(self.img_idx))
        with open(self.config_path, "w") as configfile:
            self.config.write(configfile)

    def __create_print_statements(self,
                                  frame: Union[int, Tuple[int, int]]):

        for target in self.clf_names:
            target_present_choice = self.clf_cbs[target].get()
            if target_present_choice:
                if isinstance(frame, int):
                    print(f"{target} PRESENT in frame {frame}")
                else:
                    print(f"{target} PRESENT in frames {frame[0]} to {frame[1]}")
            else:
                if isinstance(frame, int):
                    print(f"{target} ABSENT in frame {frame}")
                else:
                    print(f"{target} ABSENT in frames {frame[0]} to {frame[1]}")

def select_labelling_video(config_path: Union[str, os.PathLike],
                           threshold_dict: Optional[Dict[str, Entry_Box]] = None,
                           setting: str = None,
                           continuing: bool = None,
                           video_file_path: Union[str, os.PathLike] = None):

    check_file_exist_and_readable(file_path=config_path)
    if threshold_dict is not None:
        check_valid_dict(x=threshold_dict, valid_key_dtypes=(str,), valid_values_dtypes=(float,))
    if setting is not None:
        check_str(name='setting', value=setting, options=('pseudo', "from_scratch",))
    check_valid_boolean(value=[continuing], source=select_labelling_video.__name__)

    config = read_config_file(config_path=config_path)
    project_path = read_config_entry(config , ConfigKey.GENERAL_SETTINGS.value, ConfigKey.PROJECT_PATH.value,data_type=Dtypes.STR.value)
    video_dir = os.path.join(project_path, 'videos')
    video_dir = None if not os.path.isdir(video_dir) else video_dir

    if setting is not "pseudo":
        video_file_path = filedialog.askopenfilename(filetypes=[("Video files", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)], initialdir=video_dir)
    else:
        threshold_dict_values = {}
        for k, v in threshold_dict.items():
            check_float(name=k, value=float(v.entry_get), min_value=0.0, max_value=1.0)
            threshold_dict_values[k] = float(v.entry_get)
        threshold_dict = threshold_dict_values

    check_file_exist_and_readable(file_path=video_file_path)
    video_meta = get_video_meta_data(video_file_path)
    _, video_name, _ = get_fn_ext(video_file_path)



    print(f"ANNOTATING VIDEO {video_name} \n VIDEO INFO: {video_meta}")
    _ = LabellingInterface(config_path=config_path,
                           file_path=video_file_path,
                           threshold_dict=threshold_dict,
                           setting=setting,
                           continuing=continuing)


# test = select_labelling_video(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                               threshold_dict=None, #threshold_dict={'Attack': 0.4}
#                               setting='from_scratch',
#                               continuing=False)

# _ = LabellingInterface(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini",
#                        file_path=r"C:\troubleshooting\mitra\project_folder\videos\501_MA142_Gi_CNO_0521.mp4",
#                        threshold_dict=None,
#                        setting='from_scratch',
#                        continuing=False)





# #
# test = select_labelling_video(config_path=r"/Users/simon/Desktop/envs/simba/troubleshooting/mitra/project_folder/project_config.ini",
#                               threshold_dict={'Attack': 0.4},
#                               setting='from_scratch',
#                               continuing=False)



__author__ = "Simon Nilsson"

from tkinter import *

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.transform.simba_to_yolo import \
    SimBA2Yolo
from simba.third_party_label_appenders.transform.utils import (
    get_yolo_keypoint_bp_id_idx, get_yolo_keypoint_flip_idx)
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FileSelect,
                                        FolderSelect, SimBADropDown)
from simba.utils.read_write import str_2_bool

TRAIN_SIZE_OPTIONS = np.arange(10, 110, 10)
SAMPLE_SIZE_OPTIONS = list(np.arange(50, 650, 50))

THRESHOLD_OPTIONS = list(np.arange(0.1, 1.1, 0.1))

PADDING_OPTIONS = list(np.round(np.arange(0.01, 10.05, 0.05),2).astype(str))
PADDING_OPTIONS = list(np.insert(PADDING_OPTIONS, 0, 'None'))

class SimBA2YoloKeypointsPopUp(PopUpMixin):

    """
    :example:
    >>> SimBA2YoloKeypointsPopUp()
    """
    def __init__(self):
        PopUpMixin.__init__(self, title="SIMBA TO YOLO KEYPOINTS", icon='SimBA_logo_3_small')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.config_select = FileSelect(parent=settings_frm, fileDescription='SIMBA PROJECT CONFIG (.INI): ', lblwidth=35, file_types=[("INI FILE", (".ini", ".INI",))], entry_width=40, initialdir=r"C:\troubleshooting\mitra\project_folder")
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, entry_width=40, initialdir=r'C:\troubleshooting\mitra\yolo')
        self.train_size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=TRAIN_SIZE_OPTIONS, label="TRAIN SIZE (%): ", label_width=35, dropdown_width=40, value=70)
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE: ", label_width=35, dropdown_width=40, value='TRUE')
        self.padding_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=PADDING_OPTIONS, label="PADDING: ", label_width=35, dropdown_width=40, value='None')
        self.sample_size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=SAMPLE_SIZE_OPTIONS, label="FRAME SAMPLES PER VIDEO: ", label_width=35, dropdown_width=40, value=100)
        self.threshold_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=THRESHOLD_OPTIONS, label="CONFIDENCE THRESHOLD: ", label_width=35, dropdown_width=40, value='None')
        self.grey_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.clahe_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="CLAHE: ", label_width=35, dropdown_width=40, value='FALSE')

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.config_select.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.train_size_dropdown.grid(row=2, column=0, sticky=NW)

        self.verbose_dropdown.grid(row=3, column=0, sticky=NW)
        self.padding_dropdown.grid(row=4, column=0, sticky=NW)
        self.sample_size_dropdown.grid(row=5, column=0, sticky=NW)
        self.threshold_dropdown.grid(row=6, column=0, sticky=NW)
        self.grey_dropdown.grid(row=7, column=0, sticky=NW)
        self.clahe_dropdown.grid(row=8, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()


    def run(self):
        config_path = self.config_select.file_path
        config = ConfigReader(config_path=config_path)
        animal_names = list(config.animal_bp_dict.keys())
        bps = [x[:-2] for x in config.animal_bp_dict[animal_names[0]]['X_bps']]
        flip_idx = get_yolo_keypoint_flip_idx(x=bps)
        map_dict = {c: k for c, k in enumerate(animal_names)}
        bp_id_idx = None
        if len(animal_names) > 1:
            bp_id_idx = get_yolo_keypoint_bp_id_idx(animal_bp_dict=config.animal_bp_dict)
        train_size = int(self.train_size_dropdown.get_value()) / 100
        verbose = str_2_bool(self.verbose_dropdown.get_value())
        save_dir = self.save_dir.folder_path
        padding = float(self.padding_dropdown.get_value()) if self.padding_dropdown.get_value() != 'None' else 0.0
        sample_size = int(self.sample_size_dropdown.get_value())
        grey = str_2_bool(self.grey_dropdown.get_value())
        clahe = str_2_bool(self.clahe_dropdown.get_value())
        threshold = float(self.threshold_dropdown.get_value())

        runner = SimBA2Yolo(config_path=config_path,
                            save_dir=save_dir,
                            data_dir=None,
                            train_size=train_size,
                            threshold=threshold,
                            verbose=verbose,
                            greyscale=grey,
                            padding=padding,
                            flip_idx=flip_idx,
                            names=map_dict,
                            sample_size=sample_size,
                            bp_id_idx=bp_id_idx,
                            clahe=clahe)
        runner.run()


#SimBA2YoloKeypointsPopUp()
#simba_to_yolo_keypoints(config_path=r"C:\troubleshooting\mitra\project_folder\project_config.ini", save_dir=r'C:\troubleshooting\mitra\yolo', sample_size=150, verbose=True)



#### MODIFIED FROM @Toshea111 - https://github.com/Toshea111/sleap/blob/develop/docs/notebooks/Convert_HDF5_to_CSV_updated.ipynb
import io
import os
from typing import Any, Dict, List, Union

import h5py
import pandas as pd

from simba.data_processors.interpolate import Interpolate
from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pose_importer_mixin import PoseImporterMixin
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_int,
                                check_str, check_valid_lst)
from simba.utils.enums import Methods, TagNames
from simba.utils.errors import BodypartColumnNotFoundError
from simba.utils.printing import (SimbaTimer, log_event, stdout_success,
                                  stdout_warning)
from simba.utils.read_write import (clean_sleap_file_name,
                                    find_all_videos_in_project, get_fn_ext,
                                    get_video_meta_data, write_df)


class SLEAPImporterH5(ConfigReader, PoseImporterMixin):
    """
    Importing SLEAP pose-estimation data into SimBA project in ``H5`` format

    :param str config_path: path to SimBA project config file in Configparser format
    :param str data_folder: Path to folder containing SLEAP data in `H5` format.
    :param List[str] id_lst: Animal names. This will be ignored in one animal projects and default to ``Animal_1``.
    :param Optional[Dict[str, str]] interpolation_setting: Dict defining the type and method to use to perform interpolation {'type': 'animals', 'method': 'linear'}.
    :param Optional[Dict[str, Union[str, int]]] smoothing_settings: Dictionary defining the pose estimation smoothing method {'time_window': 500, 'method': 'gaussian'}.


    .. note::
       `Multi-animal import tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Multi_animal_pose.md>`__.

    References
    ----------
    .. [1] Pereira et al., SLEAP: A deep learning system for multi-animal pose tracking, `Nature Methods`,
           2022.

    >>> sleap_h5_importer = SLEAPImporterH5(config_path=r'project_folder/project_config.ini', data_folder=r'data_folder', actor_IDs=['Termite_1', 'Termite_2', 'Termite_3', 'Termite_4', 'Termite_5'], interpolation_settings="Body-parts: Nearest", smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
    >>> sleap_h5_importer.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_folder: Union[str, os.PathLike],
                 id_lst: List[str],
                 interpolation_settings: Union[str, None],
                 smoothing_settings: Union[Dict[str, Any], None]):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PoseImporterMixin.__init__(self)
        log_event(logger_name=str(__class__.__name__),log_type=TagNames.CLASS_INIT.value, msg=self.create_log_msg_from_init_args(locals=locals()))
        check_file_exist_and_readable(file_path=config_path)
        check_if_dir_exists(in_dir=data_folder, source=self.__class__.__name__)
        check_valid_lst(data=id_lst, source=f'{self.__class__.__name__} id_lst', valid_dtypes=(str,), min_len=1)
        if interpolation_settings is not None:
            check_if_keys_exist_in_dict(data=interpolation_settings, key=['method', 'type'], name=f'{self.__class__.__name__} interpolation_settings')
            check_str(name=f'{self.__class__.__name__} interpolation_settings type', value=interpolation_settings['type'], options=('body-parts', 'animals'))
            check_str(name=f'{self.__class__.__name__} interpolation_settings method', value=interpolation_settings['method'], options=('linear', 'quadratic', 'nearest'))
        if smoothing_settings is not None:
            check_if_keys_exist_in_dict(data=smoothing_settings, key=['method', 'time_window'], name=f'{self.__class__.__name__} smoothing_settings')
            check_str(name=f'{self.__class__.__name__} smoothing_settings method', value=smoothing_settings['method'], options=('savitzky-golay', 'gaussian'))
            check_int(name=f'{self.__class__.__name__} smoothing_settings time_window', value=smoothing_settings['time_window'], min_value=1)

        self.interpolation_settings, self.smoothing_settings = interpolation_settings, smoothing_settings
        self.data_folder, self.id_lst = data_folder, id_lst
        self.import_log_path = os.path.join(self.logs_path, f"data_import_log_{self.datetime}.csv")
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir, raise_error=True if len(id_lst) > 1 else False)
        self.input_data_paths = self.find_data_files(dir=self.data_folder, extensions=[".h5"])
        if self.pose_setting is Methods.USER_DEFINED.value:
            self.__update_config_animal_cnt()
        if self.animal_cnt > 1:
            self.data_and_videos_lk = self.link_video_paths_to_data_paths(data_paths=self.input_data_paths, video_paths=self.video_paths, filename_cleaning_func=clean_sleap_file_name)
            self.check_multi_animal_status()
            self.animal_bp_dict = self.create_body_part_dictionary(self.multi_animal_status,self.id_lst,self.animal_cnt,self.x_cols,self.y_cols,self.p_cols,self.clr_lst)
            if self.pose_setting is Methods.USER_DEFINED.value:
                self.update_bp_headers_file(update_bp_headers=True)
        else:
            self.data_and_videos_lk = dict([(get_fn_ext(file_path)[1], {"DATA": file_path, "VIDEO": None}) for file_path in self.input_data_paths])
        log_event(logger_name=str(self.__class__.__name__), log_type="info", msg=f"Importing data files: {self.input_data_paths}. Interpolation: {interpolation_settings}, Smoothing: {smoothing_settings}")
        print(f"Importing {len(list(self.data_and_videos_lk.keys()))} file(s)...")

    def run(self):
        for file_cnt, (video_name, video_data) in enumerate(self.data_and_videos_lk.items()):
            self.output_filename = clean_sleap_file_name(filename=video_name)
            print(f"Importing {self.output_filename}...")
            video_timer = SimbaTimer(start=True)
            self.video_name = self.output_filename
            with h5py.File(video_data["DATA"], "r") as f:
                missing_keys = [x for x in ["tracks", "point_scores", "node_names", "track_names"] if not x in list(f.keys())]
                if missing_keys:
                    stdout_warning(msg=f'{video_data["DATA"]} is not a valid SLEAP H5 file. Missing keys {missing_keys} Skipping {self.output_filename}...')
                    continue
                tracks = f["tracks"][:].T
                point_scores = f["point_scores"][:].T


            csv_rows = []
            n_frames, n_nodes, _, n_tracks = tracks.shape
            for frame_ind in range(n_frames):
                csv_row = []
                for track_ind in range(n_tracks):
                    for node_ind in range(n_nodes):
                        for xyp in range(3):
                            if xyp == 0 or xyp == 1:
                                data = tracks[frame_ind, node_ind, xyp, track_ind]
                            else:
                                data = point_scores[frame_ind, node_ind, track_ind]
                            csv_row.append(f"{data:.3f}")
                csv_rows.append(" ".join(csv_row))
            csv_rows = "\n".join(csv_rows)
            self.data_df = pd.read_csv(io.StringIO(csv_rows), delim_whitespace=True, header=None).fillna(0)

            if len(self.data_df.columns) != len(self.bp_headers):
                raise BodypartColumnNotFoundError(
                    msg=f'The number of body-parts in data file {video_data["DATA"]} do not match the number of body-parts in your SimBA project. '
                    f"The number of of body-parts expected by your SimBA project is {int(len(self.bp_headers) / 3)}. "
                    f'The number of of body-parts contained in data file {video_data["DATA"]} is {int(len(self.data_df.columns) / 3)}. '
                    f"Make sure you have specified the correct number of animals and body-parts in your project.",
                    source=self.__class__.__name__)
            self.data_df.columns = self.bp_headers
            if self.animal_cnt > 1:
                self.initialize_multi_animal_ui(animal_bp_dict=self.animal_bp_dict, video_info=get_video_meta_data(video_data["VIDEO"]), data_df=self.data_df, video_path=video_data["VIDEO"])
                self.multianimal_identification()
            else:
                self.out_df = self.insert_multi_idx_columns(df=self.data_df.fillna(0))

            self.save_path = os.path.join(os.path.join(self.input_csv_dir, f"{self.output_filename}.{self.file_type}"))
            write_df(df=self.out_df, file_type=self.file_type, save_path=self.save_path, multi_idx_header=True)
            if self.interpolation_settings is not None:
                interpolator = Interpolate(config_path=self.config_path, data_path=self.save_path, type=self.interpolation_settings['type'], method=self.interpolation_settings['method'], multi_index_df_headers=True, copy_originals=False)
                interpolator.run()
            if self.smoothing_settings is not None:
                smoother = Smoothing(config_path=self.config_path, data_path=self.save_path, time_window=self.smoothing_settings['time_window'], method=self.smoothing_settings['method'], multi_index_df_headers=True, copy_originals=False)
                smoother.run()
            video_timer.stop_timer()
            stdout_success(msg=f"Video {self.output_filename} data imported...", elapsed_time=video_timer.elapsed_time_str, source=self.__class__.__name__)
        self.timer.stop_timer()
        stdout_success(msg=f"All SLEAP H5 data files imported to {self.input_csv_dir} directory", elapsed_time=self.timer.elapsed_time_str, source=self.__class__.__name__)



# test = SLEAPImporterH5(config_path=r"C:\troubleshooting\sleap_two_animals_open_field\project_folder\project_config.ini",
#                         data_folder=r"C:\Users\sroni\Downloads\slape_data_fodler_2",
#                         id_lst=['Jarryd', 'Nilsson'],
#                         interpolation_settings=None,
#                         smoothing_settings = None)
# test.run()



#
#
# test = SLEAPImporterH5(config_path=r"C:\troubleshooting\slp_single_rat\project_folder\project_config.ini",
#                         data_folder=r"C:\troubleshooting\slp_single_rat\h5",
#                         id_lst=['Jarryd'],
#                         interpolation_settings=None,
#                         smoothing_settings = None)
# test.run()



# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/simba/troubleshooting/sleap_two_animals/project_folder/project_config.ini",
#                         data_folder=r'/Users/simon/Desktop/envs/simba/troubleshooting/sleap_two_animals/h5_import',
#                         id_lst=['White', 'Black'],
#                         interpolation_settings={'type': 'body-parts', 'method': 'nearest'},
#                         smoothing_settings = {'time_window': 500, 'method': 'gaussian'})
# test.run()


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_import/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_import/data_h5',
#                    id_lst=['White', 'Black'],
#                    interpolation_settings= "Body-parts: Nearest", #'"Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms_2/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms_2/data',
#                    id_lst=['Simon'],
#                    interpolation_settings= 'None', #'"Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()

# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/import_h5',
#                    id_lst=['Simon', 'Nastacia', 'Ben', 'John', 'JJ'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()
# print('All SLEAP imports complete.')


#
# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_dropped_frms/data',
#                    id_lst=['Simon'],
#                    interpolation_settings= 'None', #'"Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'None', 'Parameters': {'Time_window': '200'}})
# test.run()


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/SLEAP_2_Animals_16_body_parts/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/SLEAP_2_Animals_16_body_parts/data/h5',
#                    id_lst=['Simon', 'Nastacia'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.run()
# print('All SLEAP imports complete.')


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/Termites_5/import_h5',
#                    id_lst=['Simon', 'Nastacia', 'Ben', 'John', 'JJ'],
#                    interpolation_settings='None',
#                    smoothing_settings = {'Method': 'None'})
# test.run()
# print('All SLEAP imports complete.')
#
# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_cody/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_cody/import_h5',
#                    actor_IDs=['Nastacia', 'Sam'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.import_sleap()
# print('All SLEAP imports complete.')


# test = SLEAPImporterH5(config_path="/Users/simon/Desktop/envs/troubleshooting/sleap_cody/project_folder/project_config.ini",
#                    data_folder=r'/Users/simon/Desktop/envs/troubleshooting/sleap_cody/import_h5',
#                    actor_IDs=['Nastacia'],
#                    interpolation_settings="Body-parts: Nearest",
#                    smoothing_settings = {'Method': 'Savitzky Golay', 'Parameters': {'Time_window': '200'}})
# test.import_sleap()
# # print('All SLEAP imports complete.')



from tkinter import ttk
from tkinter.ttk import Style, Treeview
from typing import Any, Dict, Optional, Tuple

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.utils.checks import check_str, check_valid_dict, check_valid_tuple
from simba.utils.enums import Formats


class GetTreeView(PopUpMixin):
    """
    Create a hierarchical TreeView widget for displaying nested dictionary data in a popup window.

    This class creates a TreeView interface that displays nested dictionary data in an expandable tree format.
    Each top-level dictionary key becomes a parent node, and the nested dictionary entries become child nodes.
    The TreeView includes vertical and horizontal scrollbars and supports double-click expansion/collapse functionality.

    .. note::
       The TreeView automatically calculates its height based on the window size to maximize screen usage. Parent nodes can be expanded/collapsed by double-clicking.

    .. image:: _static/img/tree_view.png
       :width: 700
       :align: center

    :param Dict[str, Dict[str, Any]] data: Nested dictionary data where top-level keys become parent nodes  and nested dictionaries become child nodes.
    :param Optional[str] index_col_name: Header text for the tree column (first column). Default is empty string.
    :param Optional[Tuple[str, ...]] headers: Tuple of column headers for data columns. Default is ("VALUE",).
    :param str title: Window title for the popup. Default is 'DATA'.
    :param Optional[Tuple[int, int]] size: Window size as (width, height) tuple. Default is (800, 500).
    :param str theme: Tkinter theme name to use. Default is 'clam' for better visual separation.

    :example:
    >>> from simba.utils.read_write import find_all_videos_in_directory
    >>> from simba.utils.read_write import get_video_meta_data
    >>> videos = find_all_videos_in_directory(directory=r'C:\troubleshooting\mitra\project_folder\videos', as_dict=True)
    >>> results = {}
    >>> for video_name, video_path in videos.items():
    >>>     v = get_video_meta_data(video_path=video_path)
    >>>     results[v['video_name']] = v
    >>> GetTreeView(data=results, index_col_name='VIDEO')
    
    :example:
    >>> # Simple example with custom data
    >>> data = {
    >>>     'Video1': {'fps': 30, 'width': 1920, 'height': 1080},
    >>>     'Video2': {'fps': 25, 'width': 1280, 'height': 720}
    >>> }
    >>> tree_view = GetTreeView(data=data, index_col_name='Videos', title='Video Metadata')
    """

    def __init__(self,
                 data: Dict[str, Dict[str, Any]],
                 index_col_name: Optional[str] = '',
                 headers: Optional[Tuple[str, ...]] = ("VALUE",),
                 title: str = 'DATA',
                 size: Optional[Tuple[int, int]] = (800, 500),
                 theme: str = 'clam'):

        check_valid_dict(x=data, valid_key_dtypes=(str,), valid_values_dtypes=(dict,), min_len_keys=1, source=f'{self.__class__.__name__} data')
        check_valid_tuple(x=headers, source=f'{self.__class__.__name__} headers', valid_dtypes=(str,))
        check_str(name=f'{self.__class__.__name__} title', value=title, allow_blank=False, raise_error=True)
        for cnt, (i, x) in enumerate(data.items()):
            check_valid_dict(x=x, valid_key_dtypes=(str,), min_len_keys=1, source=f'{self.__class__.__name__} data {i}')
        PopUpMixin.__init__(self, title=title, size=size, icon='eye')

        self.main_frm.pack(fill="both", expand=True)

        window_height = size[1] if size else 500
        tree_height = int((window_height - 50) / 20)  # Estimate rows based on window height
        self.tree = Treeview(self.main_frm, columns=headers, show="tree headings", height=tree_height)
        self.tree.pack(fill="both", expand=True, side="left")
        style = Style()
        style.theme_use(theme)

        self.tree.heading('#0', text=index_col_name, anchor="center")
        self.tree.column('#0', anchor="w", width=450, stretch=True)

        for col in headers:
            self.tree.heading(col, text=col, anchor="center")
            self.tree.column(col, anchor="w", width=320, stretch=True)

        vsb = ttk.Scrollbar(self.main_frm, orient="vertical", command=self.tree.yview)
        hsb = ttk.Scrollbar(self.main_frm, orient="horizontal", command=self.tree.xview)
        self.tree.configure(yscrollcommand=vsb.set, xscrollcommand=hsb.set)

        vsb.pack(side="right", fill="y")
        hsb.pack(side="bottom", fill="x")
        
        self.insert_data(data)
        self.tree.bind("<Double-1>", self.toggle_node)
        #self.main_frm.mainloop()

    def insert_data(self, data: Dict[str, Dict[str, Any]]):
        for parent_key, child_dict in data.items():
            parent_id = self.tree.insert("", "end", text=parent_key, values=("",), tags=('parent',), open=False)
            for k, v in child_dict.items():
                formatted_value = self.format_value(v)
                self.tree.insert(parent_id, "end", text=k, values=(formatted_value,), tags=('child',))

        self.tree.tag_configure('parent', background='#f0f0f0', font=Formats.FONT_HEADER.value)
        self.tree.tag_configure('child', background='#ffffff', font=Formats.FONT_REGULAR.value)

    def format_value(self, value: Any) -> str:
        if isinstance(value, (int, float)):
            return f"{value:,}" if isinstance(value, int) else f"{value:.2f}"
        elif isinstance(value, str):
            return value
        else:
            return str(value)

    def toggle_node(self, event):
        item = self.tree.selection()[0]
        if self.tree.parent(item):
            return
        if self.tree.item(item, "open"):
            self.tree.item(item, open=False)
        else:
            self.tree.item(item, open=True)

    def expand_all(self):
        for item in self.tree.get_children():
            self.tree.item(item, open=True)

    def collapse_all(self):
        for item in self.tree.get_children():
            self.tree.item(item, open=False)


# from simba.utils.read_write import find_all_videos_in_directory
# from simba.utils.read_write import get_video_meta_data
# videos = find_all_videos_in_directory(directory=r'D:\EPM_4\original', as_dict=True)
# results = {}
# for video_name, video_path in videos.items():
#     v = get_video_meta_data(video_path=video_path)
#     results[v['video_name']] = v
# GetTreeView(data=results, index_col_name='VIDEO')


import os
from copy import copy
from typing import Optional, Union

from simba.data_processors.interpolate import Interpolate
from simba.data_processors.smoothing import Smoothing
from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (check_if_dir_exists,
                                check_if_keys_exist_in_dict, check_int,
                                check_str, check_valid_boolean)
from simba.utils.enums import ConfigKey, Dtypes, Methods
from simba.utils.errors import InvalidInputError, NoFilesFoundError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext, read_facemap_h5, write_df)


class FaceMapImporter(ConfigReader):
    """
    :example:
    >>> r = FaceMapImporter(config_path=r"C:\troubleshooting\facemap_project\project_folder\project_config.ini", data_path=r'C:\troubleshooting\facemap_project\data', smoothing_settings={'method': 'savitzky-golay', 'time_window': 100})
    >>> r.run()
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 data_path: Union[str, os.PathLike],
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 smoothing_settings: Optional[dict] = None,
                 interpolation_settings: Optional[dict] = None,
                 verbose: Optional[bool] = True):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False, create_logger=False)
        pose_config_name = self.read_config_entry(config=self.config, section=ConfigKey.CREATE_ENSEMBLE_SETTINGS.value, option=ConfigKey.POSE_SETTING.value, default_value=None, data_type=Dtypes.STR.value).strip()
        if pose_config_name != Methods.FACEMAP.value:
            raise InvalidInputError(msg=f'The project {config_path} is not a FaceMap project. Cannot import FaceMap data to a non SimBA FaceMap project ({ConfigKey.POSE_SETTING.value}: {pose_config_name}, expected: {Methods.FACEMAP.value})', source=self.__class__.__name__)
        if os.path.isdir(data_path):
            self.data_paths = find_files_of_filetypes_in_directory(directory=data_path, extensions=['.h5'], raise_error=True)
        elif os.path.isfile(data_path):
            self.data_paths = [data_path]
        else:
            raise NoFilesFoundError(msg=f'{data_path} is not a valid file path or valid directory path', source=self.__class__.__name__)
        if interpolation_settings is not None:
            check_if_keys_exist_in_dict(data=interpolation_settings, key=['method', 'type'], name=f'{self.__class__.__name__} interpolation_settings')
            check_str(name=f'{self.__class__.__name__} interpolation_settings type', value=interpolation_settings['type'], options=('body-parts', 'animals'))
            check_str(name=f'{self.__class__.__name__} interpolation_settings method', value=interpolation_settings['method'], options=('linear', 'quadratic', 'nearest'))
            self.interpolation_type, self.interpolation_method = interpolation_settings['type'], interpolation_settings['method']
        else:
            self.interpolation_type, self.interpolation_method = None, None
        if smoothing_settings is not None:
            check_if_keys_exist_in_dict(data=smoothing_settings, key=['method', 'time_window'], name=f'{self.__class__.__name__} smoothing_settings')
            check_str(name=f'{self.__class__.__name__} smoothing_settings method', value=smoothing_settings['method'], options=('savitzky-golay', 'gaussian'))
            check_int(name=f'{self.__class__.__name__} smoothing_settings time_window', value=smoothing_settings['time_window'], min_value=1)
            self.smoothing_time, self.smoothing_method = smoothing_settings['time_window'], smoothing_settings['method']
        else:
            self.smoothing_time, self.smoothing_method = None, None
        check_valid_boolean(value=verbose, source=f'{self.__class__.__name__} verbose', raise_error=True)
        if save_dir is not None:
            check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} save_dir', raise_error=True)
        else:
            save_dir = copy(self.outlier_corrected_dir)
        self.interpolation_settings, self.smoothing_settings, = (interpolation_settings, smoothing_settings)
        self.save_dir, self.verbose = save_dir, verbose

    def run(self):
        for file_cnt, file_path in enumerate(self.data_paths):
            video_name = get_fn_ext(filepath=file_path)[1]
            if self.verbose:
                print(f'Importing FaceMap data for video {video_name}...')
            file_timer = SimbaTimer(start=True)
            data_df = read_facemap_h5(file_path=file_path)
            save_path = os.path.join(self.save_dir, f'{video_name}.csv')
            write_df(df=data_df, file_type=self.file_type, save_path=save_path, multi_idx_header=False)
            if self.interpolation_settings is not None:
                interpolator = Interpolate(config_path=self.config_path, data_path=save_path, type=self.interpolation_type, method=self.interpolation_method, multi_index_df_headers=False, copy_originals=False)
                interpolator.run()
            if self.smoothing_settings is not None:
                smoother = Smoothing(config_path=self.config_path, data_path=save_path, time_window=self.smoothing_time, method=self.smoothing_method, multi_index_df_headers=False, copy_originals=False)
                smoother.run()
            file_timer.stop_timer()
            print(f'Imported data for video {video_name} (elapsed time: {file_timer.elapsed_time}s)')
        self.timer.stop_timer()
        stdout_success(msg=f"{len(self.data_paths)} SimBA FaceMap tracking files file(s) imported to the SimBA project {self.save_dir}", source=self.__class__.__name__, elapsed_time=self.timer.elapsed_time)



# r = FaceMapImporter(config_path=r"C:\troubleshooting\facemap_project\project_folder\project_config.ini", data_path=r'C:\troubleshooting\facemap_project\data', smoothing_settings={'method': 'savitzky-golay', 'time_window': 100})
# r.run()


__author__ = "Simon Nilsson"

import os
from typing import Optional, Union

import cv2
import numpy as np

from simba.mixins.image_mixin import ImageMixin
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_instance, check_int)
from simba.utils.enums import TextOptions
from simba.utils.errors import FrameRangeError
from simba.utils.read_write import (get_fn_ext, get_video_meta_data,
                                    read_frm_of_video)

PIXEL_SENSITIVITY = 20
DRAW_COLOR = (144, 0, 255)

class GetPixelsPerMillimeterInterface():
    """
    Graphical interface to compute how many pixels represents a metric millimeter.

    .. video:: _static/img/GetPixelsPerMillimeterInterface.webm
       :width: 800
       :autoplay:
       :loop:

    :param Union[str, os.PathLike] video_path: Path to a video file on disk.
    :param float known_metric_mm: Known millimeter distance to get the pixels conversion factor for.
    :returns: The number of pixels per metric millimeter.
    :rtype: float

    :example:
    >>> runner = GetPixelsPerMillimeterInterface(video_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi', known_metric_mm=140)
    >>> runner.run()
    >>> runner = GetPixelsPerMillimeterInterface(video_path='/Users/simon/Desktop/blank_video.mp4', known_metric_mm=140)
    >>> runner.run()
    """

    def __init__(self,
                 video_path: Union[str, os.PathLike],
                 known_metric_mm: float):

        check_file_exist_and_readable(file_path=video_path)
        self.video_meta_data = get_video_meta_data(video_path=video_path)
        check_float(name='distance', value=known_metric_mm, min_value=1)
        self.video_path, self.known_metric_mm = video_path, float(known_metric_mm)
        self.frame, _ = ImageMixin.find_first_non_uniform_clr_frm(video_path=video_path, start_idx=0, end_idx=self.video_meta_data['fps'])
        self.video_dir, self.video_name, _ = get_fn_ext(filepath=self.video_path)
        self.font_scale, self.spacing_x, self.spacing_y = PlottingMixin().get_optimal_font_scales(text='"Press ESC to save and exit"', accepted_px_width=int(self.video_meta_data['width']), accepted_px_height=int(self.video_meta_data['height'] * 0.3))
        self.circle_scale = PlottingMixin().get_optimal_circle_size(frame_size=(int(self.video_meta_data['width']), int(self.video_meta_data['height'])), circle_frame_ratio=70)
        self.original_frm = self.frame.copy()
        self.overlay_frm = self.frame.copy()
        self.ix, self.iy = -1, -1
        self.cord_results = []
        self.cord_status = False
        self.move_status = False
        self.insert_status = False
        self.change_loop = False
        self.coord_change = []
        self.new_cord_lst = []

    def _draw_circle(self, event, x, y, flags, param):
        if (event == cv2.EVENT_LBUTTONDBLCLK) and (len(self.cord_results) < 4):
            cv2.circle(self.overlay_frm, (x, y), self.circle_scale, DRAW_COLOR, -1)
            self.cord_results.append(x)
            self.cord_results.append(y)
            if len(self.cord_results) == 4:
                self.cord_status = True
                cv2.line(self.overlay_frm, (self.cord_results[0], self.cord_results[1]), (self.cord_results[2], self.cord_results[3]), DRAW_COLOR, 6)

    def _select_cord_to_change(self, event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDBLCLK:
            if np.sqrt((x - self.cord_results[0]) ** 2 + (y - self.cord_results[1]) ** 2) <= PIXEL_SENSITIVITY:
                self.coord_change = [1, self.cord_results[0], self.cord_results[1]]
                self.move_status = True
            elif np.sqrt((x - self.cord_results[2]) ** 2 + (y - self.cord_results[3]) ** 2) <= PIXEL_SENSITIVITY:
                self.coord_change = [2, self.cord_results[2], self.cord_results[3]]
                self.move_status = True

    def _select_new_dot_location(self, event, x, y, flags, param):
        if event == cv2.EVENT_LBUTTONDBLCLK:
            self.new_cord_lst.append(x)
            self.new_cord_lst.append(y)
            self.insert_status = True

    def run(self):
        cv2.namedWindow("Select coordinates: double left mouse click at two locations. Press ESC when done", cv2.WINDOW_NORMAL)
        while 1:
            if self.cord_status == False and (self.move_status == False) and (self.insert_status == False):
                cv2.setMouseCallback("Select coordinates: double left mouse click at two locations. Press ESC when done", self._draw_circle)
                cv2.imshow("Select coordinates: double left mouse click at two locations. Press ESC when done", self.overlay_frm)
                k = cv2.waitKey(20) & 0xFF
                if k == 27:
                    break
            if (self.cord_status == True) and (self.move_status == False) and (self.insert_status == False):
                if self.change_loop == True:
                    self.overlay_frm = self.original_frm.copy()
                    cv2.circle(self.overlay_frm, (self.cord_results[0], self.cord_results[1]), self.circle_scale, DRAW_COLOR, -1)
                    cv2.circle(self.overlay_frm, (self.cord_results[2], self.cord_results[3]), self.circle_scale, DRAW_COLOR, -1)
                    cv2.line(self.overlay_frm, (self.cord_results[0], self.cord_results[1]), (self.cord_results[2], self.cord_results[3]), DRAW_COLOR, int(self.circle_scale / 3))
                cv2.putText(self.overlay_frm, "Click on circle to move", (TextOptions.BORDER_BUFFER_X.value, 50), cv2.FONT_HERSHEY_TRIPLEX, self.font_scale, DRAW_COLOR, 2)
                cv2.putText(self.overlay_frm, "Press ESC to save and exit", (TextOptions.BORDER_BUFFER_X.value, 50 + self.spacing_y), cv2.FONT_HERSHEY_TRIPLEX, self.font_scale, DRAW_COLOR, 2)
                cv2.imshow("Select coordinates: double left mouse click at two locations. Press ESC when done", self.overlay_frm)
                cv2.setMouseCallback("Select coordinates: double left mouse click at two locations. Press ESC when done", self._select_cord_to_change)

            if (self.move_status == True) and (self.insert_status == False):
                if self.change_loop == True:
                    self.frame = self.original_frm.copy()
                    self.change_loop = False
                if self.coord_change[0] == 1:
                    cv2.circle(self.frame, (self.cord_results[2], self.cord_results[3]), self.circle_scale, DRAW_COLOR, -1)
                if self.coord_change[0] == 2:
                    cv2.circle(self.frame, (self.cord_results[0], self.cord_results[1]), self.circle_scale, DRAW_COLOR, -1)
                cv2.imshow("Select coordinates: double left mouse click at two locations. Press ESC when done", self.frame)
                cv2.putText(self.frame, "Click on new circle location", (TextOptions.BORDER_BUFFER_X.value, 50), cv2.FONT_HERSHEY_TRIPLEX, self.font_scale, DRAW_COLOR, 2)
                cv2.setMouseCallback("Select coordinates: double left mouse click at two locations. Press ESC when done", self._select_new_dot_location)

            if self.insert_status == True:
                if self.coord_change[0] == 1:
                    cv2.circle(self.frame, (self.cord_results[2], self.cord_results[3]), self.circle_scale, DRAW_COLOR, -1)
                    cv2.circle(self.frame, (self.new_cord_lst[-2], self.new_cord_lst[-1]), self.circle_scale, DRAW_COLOR, -1)
                    cv2.line(self.frame, (self.cord_results[2], self.cord_results[3]), (self.new_cord_lst[-2], self.new_cord_lst[-1]), DRAW_COLOR, int(self.circle_scale / 3))
                    self.cord_results = [self.new_cord_lst[-2], self.new_cord_lst[-1], self.cord_results[2], self.cord_results[3]]
                    self.cord_status = True
                    self.move_status = False
                    self.insert_status = False
                    self.change_loop = True
                if self.coord_change[0] == 2:
                    cv2.circle(self.frame, (self.cord_results[0], self.cord_results[1]), self.circle_scale, DRAW_COLOR, -1)
                    cv2.circle(self.frame, (self.new_cord_lst[-2], self.new_cord_lst[-1]), self.circle_scale, DRAW_COLOR, -1)
                    cv2.line(self.frame, (self.cord_results[0], self.cord_results[1]), (self.new_cord_lst[-2], self.new_cord_lst[-1]), DRAW_COLOR, int(self.circle_scale / 3))
                    self.cord_results = [self.cord_results[0], self.cord_results[1], self.new_cord_lst[-2], self.new_cord_lst[-1]]
                    self.cord_status = True
                    self.move_status = False
                    self.insert_status = False
                    self.change_loop = True
                cv2.imshow("Select coordinates: double left mouse click at two locations. Press ESC when done", self.frame)
            k = cv2.waitKey(20) & 0xFF
            if k == 27:
                break

        self.cord_status = False
        self.move_status = False
        self.insert_status = False
        self.change_loop = False
        cv2.destroyAllWindows()
        euclidean_px_dist = np.sqrt((self.cord_results[0] - self.cord_results[2]) ** 2 + (self.cord_results[1] - self.cord_results[3]) ** 2)
        self.ppm = euclidean_px_dist / self.known_metric_mm


import os
from typing import Optional, Union

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import (
    check_all_file_names_are_represented_in_video_log,
    check_if_df_field_is_boolean, check_that_column_exist, check_valid_boolean)
from simba.utils.data import detect_bouts
from simba.utils.errors import CountError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class AnnotationMetaDataExtractor(ConfigReader):

    """
    Extract annotation statistics (number of annotated frames, seconds, bouts etc.) for all classifiers in a SimBA project
    to MS Excel format.

    .. note::
       `Example expected output <https://github.com/sgoldenlab/simba/blob/master/misc/ANNOTATION_STATISTICS_20240713132805.xlsx>`__.

    :param Union[str, os.PathLike] config_path: path to SimBA configparser.ConfigParser project_config.ini
    :param Optional[bool] annotated_bouts: If True, includes information on annotated bouts (start and stop time and bout length). Default True.
    :param Optional[bool] split_by_video: If True, includes a worksheet where the annotation counts are split by video. Default True.

    :example:
    >>> annotation_meta_extractor = AnnotationMetaDataExtractor(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
    >>> annotation_meta_extractor.run()
    >>> annotation_meta_extractor.save()

    """
    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 split_by_video: Optional[bool] = True,
                 annotated_bouts: Optional[bool] = True):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=True)
        if len(self.clf_names) == 0:
            raise CountError(msg=f'No classifier names associated with SimBA project {config_path}', source=self.__class__.__name__)
        if len(self.target_file_paths) == 0:
            raise CountError(msg=f'No data files found inside the {self.targets_folder} directory', source=self.__class__.__name__)
        self.save_path = os.path.join(self.logs_path, f'ANNOTATION_STATISTICS_{self.datetime}.xlsx')
        check_valid_boolean(value=[annotated_bouts, split_by_video])
        self.annotated_bouts, self.split_by_video = annotated_bouts, split_by_video

    def run(self):
        check_all_file_names_are_represented_in_video_log(video_info_df=self.video_info_df, data_paths=self.target_file_paths)
        self.results, self.bout_data = {}, []
        print(f'Analyzing annotations in {len(self.target_file_paths)} data file(s)...')
        for file_cnt, file_path in enumerate(self.target_file_paths):
            _, video_name, _ = get_fn_ext(filepath=file_path)
            self.results[video_name] = {}
            _, _, fps = self.read_video_info(video_name=video_name)
            print(f'Analyzing annotations in {video_name }... ')
            df = read_df(file_path=file_path, file_type=self.file_type)
            check_that_column_exist(df=df, column_name=self.clf_names, file_name=file_path)
            bouts = detect_bouts(data_df=df, target_lst=self.clf_names, fps=fps)
            bouts.columns = ['ANNOTATED CLASSIFIER', 'ANNOTATED START TIME', 'ANNOTATED END TIME', 'ANNOTATED START FRAME','ANNOTATED END FRAME','ANNOTATED BOUT TIME (S)']
            bouts['VIDEO'] = video_name
            bouts = bouts[['VIDEO', 'ANNOTATED CLASSIFIER', 'ANNOTATED START TIME', 'ANNOTATED END TIME', 'ANNOTATED START FRAME','ANNOTATED END FRAME','ANNOTATED BOUT TIME (S)']]
            self.bout_data.append(bouts)
            for clf in self.clf_names:
                check_if_df_field_is_boolean(df=df, field=clf, df_name=file_path)
                present_df, absent_df = df[df[clf] == 1], df[df[clf] == 0]
                self.results[video_name][clf] = {f'ANNOTATED PRESENT FRAME COUNT': len(present_df),
                                                 f'ANNOTATED PRESENT TIME (S)': round((len(present_df) / fps), 4),
                                                 f'ANNOTATED ABSENT FRAMES COUNT': len(absent_df),
                                                 f'ANNOTATED ABSENT TIME (S)': round((len(absent_df) / fps), 4)}


    def __aggregates(self):
        self.by_video = pd.DataFrame(columns=['VIDEO', 'CLASSIFIER', 'ANNOTATION MEASUREMENT', 'ANNOTATION STATISTIC'])
        for video_name, video_data in self.results.items():
            for clf in self.clf_names:
                for clf_key, clf_data in self.results[video_name][clf].items():
                    self.by_video.loc[len(self.by_video)] = [video_name, clf,  clf_key, clf_data]
        self.aggregates = pd.DataFrame(self.by_video.drop(['VIDEO'], axis=1).groupby(by=['CLASSIFIER', 'ANNOTATION MEASUREMENT'])['ANNOTATION STATISTIC'].sum())
        self.by_video = self.by_video.set_index(['VIDEO', 'CLASSIFIER'])

    def save(self):
        self.__aggregates()
        self.bout_data = pd.concat(self.bout_data, axis=0)
        with pd.ExcelWriter(self.save_path) as writer:
            self.aggregates.to_excel(writer, sheet_name='TOTAL ANNOTATION COUNTS', index=True)
            if self.split_by_video:
                self.by_video.to_excel(writer, sheet_name='VIDEO ANNOTATION COUNTS', index=True)
            if self.annotated_bouts:
                self.bout_data.to_excel(writer, sheet_name='VIDEO ANNOTATION BOUT DATA', index=False)
        self.timer.stop_timer()
        stdout_success(msg=f'Annotation data for {len(self.target_file_paths)} video(s) saved at {self.save_path}', source=self.__class__.__name__, elapsed_time=self.timer.elapsed_time)


# annotation_meta_extractor = AnnotationMetaDataExtractor(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# annotation_meta_extractor.run()
# annotation_meta_extractor.save()


import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, SimbaButton, SimbaCheckbox)
from simba.utils.checks import check_float
from simba.utils.enums import ConfigKey, Formats, Keys, Links


class ROIAnalysisPopUp(ConfigReader, PopUpMixin):

    """
    :example:
    >>> _ = ROIAnalysisPopUp(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
    """
    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        PopUpMixin.__init__(self, title="ROI ANALYSIS", size=(400, 400), icon='shapes_small')
        self.animal_cnt_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SELECT NUMBER OF ANIMALS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.ROI_DATA_ANALYSIS.value)
        self.animal_cnt_dropdown = DropDownMenu(self.animal_cnt_frm, "# of animals", list(range(1, self.animal_cnt + 1)), labelwidth=20)
        self.animal_cnt_dropdown.setChoices(1)
        self.animal_cnt_confirm_btn = SimbaButton(parent=self.animal_cnt_frm, txt="CONFIRM", img='tick', txt_clr="blue", font=Formats.FONT_REGULAR.value, cmd=self.create_settings_frm)
        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_confirm_btn.grid(row=0, column=1, sticky=NW)
        self.main_frm.mainloop()

    def create_settings_frm(self):
        if hasattr(self, "setting_frm"):
            self.setting_frm.destroy()
            self.body_part_frm.destroy()

        self.setting_frm = LabelFrame(
            self.main_frm, text="SETTINGS", font=Formats.FONT_HEADER.value
        )
        self.choose_bp_frm(parent=self.setting_frm, bp_options=self.body_parts_lst)
        self.choose_bp_threshold_frm(parent=self.setting_frm)
        self.calculate_distances_frm = LabelFrame( self.setting_frm, text="CALCULATE DISTANCES", font=Formats.FONT_HEADER.value)


        self.calculate_distance_moved_cb, self.calculate_distance_moved_var = SimbaCheckbox(parent=self.calculate_distances_frm, txt='COMPUTE DISTANCES MOVED IN ROIs', txt_img='path')
        self.calculate_distance_moved_cb.grid(row=0, column=0, sticky=NW)
        self.detailed_roi_frm = LabelFrame(self.setting_frm, text="DETAILED ROI BOUT DATA", font=Formats.FONT_HEADER.value)
        self.detailed_roi_cb, self.detailed_roi_var = SimbaCheckbox(parent=self.detailed_roi_frm, txt='DETAILED ROI BOUT DATA', txt_img='details')
        self.detailed_roi_cb.grid(row=1, column=0, sticky=NW)
        self.calculate_distances_frm.grid(row=self.frame_children(frame=self.setting_frm), column=0, sticky=NW)
        self.detailed_roi_frm.grid(row=self.frame_children(frame=self.setting_frm) + 1, column=0, sticky=NW)
        self.setting_frm.grid(row=1, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

        self.main_frm.mainloop()

    def run(self):
        settings = {}
        check_float(
            name="Probability threshold",
            value=self.probability_entry.entry_get,
            min_value=0.00,
            max_value=1.00,
        )
        settings["threshold"] = float(self.probability_entry.entry_get)
        body_parts = []
        for cnt, dropdown in self.body_parts_dropdowns.items():
            body_parts.append(dropdown.getChoices())
        roi_analyzer = ROIAnalyzer(
            config_path=self.config_path,
            data_path=None,
            calculate_distances=self.calculate_distance_moved_var.get(),
            detailed_bout_data=self.detailed_roi_var.get(),
            threshold=float(self.probability_entry.entry_get),
            body_parts=body_parts,
        )
        roi_analyzer.run()
        roi_analyzer.save()
        self.root.destroy()

# ROIAnalysisPopUp(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini")

# ROIAnalysisPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini')



import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.roi_ui import ROI_ui
from simba.roi_tools.roi_utils import multiply_ROIs, reset_video_ROIs
from simba.ui.pop_ups.min_max_draw_size_popup import SetMinMaxDrawWindowSize
from simba.ui.pop_ups.roi_size_standardizer_popup import \
    ROISizeStandardizerPopUp
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, Entry_Box,
                                        SimbaButton, SimBALabel)
from simba.utils.enums import Keys, Links
from simba.utils.errors import NoFilesFoundError
from simba.utils.read_write import find_all_videos_in_directory, get_fn_ext


class ROIVideoTable(ConfigReader, PopUpMixin):

    """
    Crates a tkinter video table listing all videos in project together with buttons associated with ROI drawing, deletaing, and duplicating shapes.

    :example:
    >>> ROIVideoTable(config_path=r"C:\troubleshooting\mouse_open_field\project_folder\project_config.ini")
    """

    def __init__(self,
                 config_path: Union[str, os.PathLike]):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False, create_logger=False)
        PopUpMixin.__init__(self, title="PROJECT VIDEOS: ROI TABLE", size=(720, 960), icon='data_table')
        self.video_dict = find_all_videos_in_directory(directory=self.video_dir, as_dict=True, raise_error=False)
        if len(list(self.video_dict.keys())) == 0:
            raise NoFilesFoundError(msg=f'Cannot draw ROIs: the {self.video_dir} directory does not contain any video files.', source=self.__class__.__name__)
        self.get_file_menu()
        self.run()

    def get_file_menu(self):
        menu = Menu(self.root)
        file_menu = Menu(menu)
        menu.add_cascade(label="File...", menu=file_menu)
        file_menu.add_command(label="Standardize ROI sizes by metric conversion factor...", compound="left", image=self.menu_icons["settings"]["img"], command=lambda: ROISizeStandardizerPopUp(config_path=self.config_path))
        file_menu.add_command(label="Set min/max draw window size...", compound="left", image=self.menu_icons["monitor"]["img"], command=lambda: SetMinMaxDrawWindowSize(config_path=self.config_path))
        self.root.config(menu=menu)

    def run(self):
        video_names = list(self.video_dict.keys())
        max_video_name_len = len(max(video_names, key=len))
        self.open_roi_uis = []
        self.video_table = CreateLabelFrameWithIcon(parent=self.main_frm, header="VIDEOS", icon_name=Keys.DOCUMENTATION.value, icon_link=Links.ROI.value)
        for table_idx in range(len(video_names)):
             _ = ROITableRow(config_path=self.config_path, parent=self.video_table, video_path=self.video_dict[video_names[table_idx]], str_width=max_video_name_len, row_idx=table_idx)
        self.video_table.grid(row=0, column=0, sticky=NW)
        self.main_frm.mainloop()

class ROITableRow():

    """
    Helper to create a single row class in the :func:`simba.ui.pop_ups.roi_video_table_pop_up.ROIVideoTable` table class.
    """
    def __init__(self,
                 config_path: Union[str, os.PathLike],
                 parent: Union[Frame, Canvas, LabelFrame, Toplevel],
                 video_path: Union[str, os.PathLike],
                 str_width: int,
                 row_idx: int):

        self.index = Entry_Box(parent=parent, fileDescription='', labelwidth=0, entry_box_width=4)
        self.index.entry_set(val=f'{row_idx+1}.')
        self.index.grid(row=row_idx, column=0, sticky=NW)
        _, video_name, _ = get_fn_ext(filepath=video_path)
        self.video_lbl = SimBALabel(parent=parent, txt=video_name, txt_clr='black', width=str_width)
        self.video_lbl.grid(row=row_idx, column=1, sticky=NW)
        self.draw_btn = SimbaButton(parent=parent, txt='DRAW', img='paint', cmd=self.draw, cmd_kwargs={'config_path': lambda: config_path, 'video_path': lambda: video_path})
        self.draw_btn.grid(row=row_idx, column=2, sticky=NW)
        self.reset_btn = SimbaButton(parent=parent, txt='RESET', img='trash', cmd=self.reset, cmd_kwargs={'config_path': lambda: config_path, 'video_path': lambda: video_path})
        self.reset_btn.grid(row=row_idx, column=3, sticky=NW)
        self.apply_to_all_btn = SimbaButton(parent=parent, txt='APPLY TO ALL', img='add_on', cmd=self.apply_to_all, cmd_kwargs={'config_path': lambda: config_path, 'video_path': lambda: video_path})
        self.apply_to_all_btn.grid(row=row_idx, column=4, sticky=NW)


    def draw(self, config_path: Union[str, os.PathLike], video_path: Union[str, os.PathLike]):
        ROI_ui(config_path=config_path, video_path=video_path)

    def reset(self, config_path: Union[str, os.PathLike], video_path: Union[str, os.PathLike]):
        reset_video_ROIs(config_path=config_path, filename=video_path)

    def apply_to_all(self, config_path: Union[str, os.PathLike], video_path: Union[str, os.PathLike]):
        multiply_ROIs(config_path=config_path, filename=video_path)


#ROIVideoTable(config_path=r"C:\troubleshooting\mouse_open_field\project_folder\project_config.ini")


import os
from tkinter import *
from tkinter import ttk
from typing import Optional, Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FileSelect,
                                        FolderSelect, SimbaButton,
                                        SimbaCheckbox, SimBADropDown)
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists, check_instance, check_int)
from simba.utils.enums import Formats, Options
from simba.utils.errors import InvalidInputError
from simba.utils.read_write import (copy_multiple_videos_to_project,
                                    copy_single_video_to_project)

#from simba.sandbox.tkinter_drag_and_drop import FileDropBox


class ImportVideosFrame(PopUpMixin, ConfigReader):

    """
    .. image:: _static/img/ImportVideosFrame.webp
       :width: 500
       :align: center

    :param Optional[Union[Frame, Canvas, LabelFrame, ttk.Frame]] parent_frm: Parent frame to insert the Import Videos frame into. If None, one is created.
    :param Optional[Union[str, os.PathLike]] config_path:
    :param Optional[int] idx_row: The row in parent_frm to insert the Videos frame into. Default: 0.
    :param Optional[int] idx_column: The column in parent_frm to insert the Videos frame into. Default: 0.

    :example:
    >>> ImportVideosFrame(config_path='/Users/simon/Desktop/envs/simba/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
    """

    def __init__(self,
                 parent_frm: Optional[Union[Frame, Canvas, LabelFrame, ttk.Frame]] = None,
                 config_path: Optional[Union[str, os.PathLike]] = None,
                 idx_row: Optional[int] = 0,
                 idx_column: Optional[int] = 0):

        if parent_frm is None and config_path is None:
            raise InvalidInputError(msg='If parent_frm is None, please pass config_path', source=self.__class__.__name__)

        elif parent_frm is None and config_path is not None:
            PopUpMixin.__init__(self, config_path=config_path, title='IMPORT VIDEO FILES', icon='import')
            parent_frm = self.main_frm

        check_instance(source=f'{ImportVideosFrame} parent_frm', accepted_types=(Frame, Canvas, LabelFrame, ttk.Frame), instance=parent_frm)
        check_int(name=f'{ImportVideosFrame} idx_row', value=idx_row, min_value=0)
        check_int(name=f'{ImportVideosFrame} idx_column', value=idx_column, min_value=0)

        import_videos_frm = CreateLabelFrameWithIcon(parent=parent_frm, header="IMPORT VIDEOS", icon_name='import', relief='solid', padx=5, pady=5)
        if config_path is None:
            Label(import_videos_frm, text="Please CREATE PROJECT CONFIG before importing VIDEOS \n", font=Formats.FONT_REGULAR.value).grid(row=0, column=0, sticky=NW)
            import_videos_frm.grid(row=0, column=0, sticky=NW)
        else:
            ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
            import_multiple_videos_frm = CreateLabelFrameWithIcon(parent=import_videos_frm, header="IMPORT MULTIPLE VIDEOS", icon_name='stack')
            self.video_directory_select = FolderSelect(import_multiple_videos_frm, "VIDEO DIRECTORY: ", lblwidth=25, entry_width=25)

            self.video_type = SimBADropDown(parent=import_multiple_videos_frm, dropdown_options=Options.VIDEO_FORMAT_OPTIONS.value, label="VIDEO FILE FORMAT: ", label_width=25, dropdown_width=25, value=Options.VIDEO_FORMAT_OPTIONS.value[0])
            import_multiple_btn = SimbaButton(parent=import_multiple_videos_frm, txt="Import MULTIPLE videos", txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.__run_video_import, cmd_kwargs={"multiple_videos": lambda: True})

            multiple_videos_symlink_cb, self.multiple_videos_symlink_var = SimbaCheckbox(parent=import_multiple_videos_frm, txt="Import SYMLINKS", txt_img='link')

            import_single_frm = CreateLabelFrameWithIcon(parent=import_videos_frm, header="IMPORT SINGLE VIDEO", icon_name='video')
            self.video_file_select = FileSelect(import_single_frm, "VIDEO PATH: ", title="Select a video file", lblwidth=25, file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)], entry_width=25)
            import_single_btn = SimbaButton(parent=import_single_frm, txt="Import SINGLE video", txt_clr='blue', font=Formats.FONT_REGULAR.value, cmd=self.__run_video_import, cmd_kwargs={"multiple_videos": lambda: False})
            single_video_symlink_cb, self.single_video_symlink_var = SimbaCheckbox(parent=import_single_frm, txt="Import SYMLINKS", txt_img='link')

            import_videos_frm.grid(row=0, column=0, sticky=NW)
            import_multiple_videos_frm.grid(row=0, sticky=W)
            self.video_directory_select.grid(row=1, sticky=W)
            self.video_type.grid(row=2, sticky=W)
            multiple_videos_symlink_cb.grid(row=3, sticky=W)
            import_multiple_btn.grid(row=4, sticky=W)

            import_single_frm.grid(row=1, column=0, sticky=NW)
            self.video_file_select.grid(row=0, sticky=W)
            single_video_symlink_cb.grid(row=1, sticky=W)
            import_single_btn.grid(row=2, sticky=W)
            import_videos_frm.grid(row=idx_row, column=idx_column, sticky=NW)

            #video_drop_box = FileDropBox(parent=import_videos_frm)
            #video_drop_box.drop_box_frm.grid(row=2, column=1, sticky=NW)



        #parent_frm.mainloop()

    def __run_video_import(self, multiple_videos: bool):
        if multiple_videos:
            check_if_dir_exists(in_dir=self.video_directory_select.folder_path)
            copy_multiple_videos_to_project(config_path=self.config_path,
                                            source=self.video_directory_select.folder_path,
                                            symlink=self.multiple_videos_symlink_var.get(),
                                            file_type=self.video_type.getChoices())

        else:
            check_file_exist_and_readable(file_path=self.video_file_select.file_path)
            copy_single_video_to_project(simba_ini_path=self.config_path,
                                         symlink=self.single_video_symlink_var.get(),
                                         source_path=self.video_file_select.file_path)


#ImportVideosFrame(config_path=r"C:\troubleshooting\RAT_NOR\project_folder\project_config.ini")


__author__ = "Simon Nilsson"

from tkinter import *

import numpy as np
import pandas as pd

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.transform.dlc_to_yolo import DLC2Yolo
from simba.third_party_label_appenders.transform.utils import (
    get_yolo_keypoint_bp_id_idx, get_yolo_keypoint_flip_idx)
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimBADropDown)
from simba.utils.checks import check_if_dir_exists
from simba.utils.read_write import recursive_file_search, str_2_bool

TRAIN_SIZE_OPTIONS = np.arange(10, 110, 10)
SAMPLE_SIZE_OPTIONS = list(np.arange(50, 650, 50))

PADDING_OPTIONS = list(np.round(np.arange(0.01, 10.05, 0.02),2).astype(str))
PADDING_OPTIONS = list(np.insert(PADDING_OPTIONS, 0, 'None'))

class DLCYoloKeypointsPopUp(PopUpMixin):

    """
    GUI for converting DeepLabCut (DLC) annotations to YOLOv8 keypoint format.

    :example:
    >>> DLCYoloKeypointsPopUp()
    """
    def __init__(self):

        PopUpMixin.__init__(self, title="DLC ANNOTATIONS TO YOLO KEY-KEYPOINTS", icon='dlc_2')
        #super().__init__(title="DLC ANNOTATIONS TO YOLO key-KEYPOINTS")
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.dlc_dir = FolderSelect(settings_frm, folderDescription="DLC DATA DIRECTORY:", lblwidth=35, entry_width=40, initialdir=r'D:\rat_resident_intruder\dlc_data') #r'D:\rat_resident_intruder\dlc_data'
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, entry_width=40, initialdir=r'D:\troubleshooting\dlc_h5_multianimal_to_yolo\yolo') #r'D:\rat_resident_intruder\yolo_1'
        self.train_size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=TRAIN_SIZE_OPTIONS, label="TRAIN SIZE (%): ", label_width=35, dropdown_width=40, value=70)
        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE: ", label_width=35, dropdown_width=40, value='TRUE')
        self.padding_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=PADDING_OPTIONS, label="PADDING: ", label_width=35, dropdown_width=40, value='None')
        self.grey_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.clahe_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="CLAHE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.animal_cnt_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=list(range(1, 10, 1)), label="ANIMAL COUNT: ", label_width=35, dropdown_width=40, value=1)

        settings_frm.grid(row=0, column=0, sticky=NW)
        self.dlc_dir.grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)
        self.train_size_dropdown.grid(row=2, column=0, sticky=NW)

        self.verbose_dropdown.grid(row=3, column=0, sticky=NW)
        self.padding_dropdown.grid(row=4, column=0, sticky=NW)
        self.grey_dropdown.grid(row=6, column=0, sticky=NW)
        self.clahe_dropdown.grid(row=7, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=8, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()


    def run(self):
        dlc_dir = self.dlc_dir.folder_path
        save_dir = self.save_dir.folder_path
        animal_cnt = int(self.animal_cnt_dropdown.get_value())
        check_if_dir_exists(in_dir=dlc_dir, source=f'{self.__class__.__name__} DLC DATA DIRECTORY', raise_error=True)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} SAVE DIRECTORY', raise_error=True)
        padding = float(self.padding_dropdown.get_value()) if self.padding_dropdown.get_value() != 'None' else 0.0
        grey = str_2_bool(self.grey_dropdown.get_value())
        clahe = str_2_bool(self.clahe_dropdown.get_value())
        train_size = int(self.train_size_dropdown.get_value()) / 100
        verbose = str_2_bool(self.verbose_dropdown.get_value())

        annotation_paths = recursive_file_search(directory=dlc_dir, substrings=['CollectedData'], extensions=['csv'], case_sensitive=False, raise_error=True)
        animal_bp_col_cnt = int(int((len(pd.read_csv(annotation_paths[0], header=[0, 1, 2]).columns) / 2) - 1) / animal_cnt)
        animal_bp_names = list(pd.read_csv(annotation_paths[0], header=[0, 1, 2]).columns)
        animal_bp_names = list(dict.fromkeys([x[1] for x in animal_bp_names[1:]]))
        flip_idx = get_yolo_keypoint_flip_idx(x=animal_bp_names[0:animal_bp_col_cnt+1])
        names = tuple([f'animal_{x+1}' for x in range(animal_cnt)])

        runner = DLC2Yolo(dlc_dir=dlc_dir, save_dir=save_dir, train_size=train_size, verbose=verbose, padding=padding, flip_idx=flip_idx, names=names, greyscale=grey, clahe=clahe)
        runner.run()


#DLCYoloKeypointsPopUp()


import math
import time
from typing import Tuple

import numpy as np
from numba import cuda

from simba.data_processors.cuda.utils import (_cuda_2d_transpose, _cuda_add_2d,
                                              _cuda_matrix_multiplication,
                                              _cuda_subtract_2d)
from simba.utils.checks import check_int, check_valid_array
from simba.utils.enums import Formats
from simba.utils.errors import InvalidInputError
from simba.utils.read_write import read_df

THREADS_PER_BLOCK = 1024


@cuda.jit()
def _egocentric_align_kernel(data, centers, rotation_vectors, results, target_angle, anchor_idx, transposed_rotation_vectors, matrix_multiplier_arr, anchor_loc):
    frm_idx = cuda.grid(1)
    if frm_idx >= data.shape[0]:
        return
    else:
        frm_points = data[frm_idx]
        frm_anchor_1, frm_anchor_2 = frm_points[anchor_idx[0]], frm_points[anchor_idx[1]]
        centers[frm_idx][0], centers[frm_idx][1] = frm_anchor_1[0], frm_anchor_1[1]
        delta_x, delta_y = frm_anchor_2[0] - frm_anchor_1[0], frm_anchor_2[1] - frm_anchor_1[1]
        frm_angle = math.atan2(delta_y, delta_x)
        frm_rotation_angle = target_angle[0] - frm_angle
        frm_cos_theta, frm_sin_theta = math.cos(frm_rotation_angle), math.sin(frm_rotation_angle)
        rotation_vectors[frm_idx][0][0], rotation_vectors[frm_idx][0][1] = frm_cos_theta, -frm_sin_theta
        rotation_vectors[frm_idx][1][0], rotation_vectors[frm_idx][1][1] = frm_sin_theta, frm_cos_theta
        keypoints_rotated = _cuda_subtract_2d(frm_points, frm_anchor_1)
        r_transposed = _cuda_2d_transpose(rotation_vectors[frm_idx], transposed_rotation_vectors[frm_idx])
        keypoints_rotated = _cuda_matrix_multiplication(keypoints_rotated, r_transposed, matrix_multiplier_arr[frm_idx])
        anchor_1_position_after_rotation = keypoints_rotated[anchor_idx[0]]
        anchor_1_position_after_rotation[0] = anchor_loc[0] - anchor_1_position_after_rotation[0]
        anchor_1_position_after_rotation[1] = anchor_loc[1] - anchor_1_position_after_rotation[1]

        frm_results = _cuda_add_2d(keypoints_rotated, anchor_1_position_after_rotation)
        for i in range(frm_results.shape[0]):
            for j in range(frm_results.shape[1]):
                results[frm_idx][i][j] = frm_results[i][j]


def egocentrically_align_pose_cuda(data: np.ndarray,
                                   anchor_1_idx: int,
                                   anchor_2_idx: int,
                                   anchor_location: np.ndarray,
                                   direction: int,
                                   batch_size: int = int(10e+5)) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

    """
    Aligns a set of 2D points egocentrically based on two anchor points and a target direction using GPU acceleration.

    Rotates and translates a 3D array of 2D points (e.g., time-series of frame-wise data) such that
    one anchor point is aligned to a specified location, and the direction between the two anchors is aligned
    to a target angle.

    .. video:: _static/img/EgocentricalAligner.webm
       :width: 600
       :autoplay:
       :loop:

    .. seealso::
       For numpy function, see :func:`simba.utils.data.egocentrically_align_pose`.
       For numba alternative, see :func:`simba.utils.data.egocentrically_align_pose_numba`.
       To align both pose and video, see :func:`simba.data_processors.egocentric_aligner.EgocentricalAligner`.
       To egocentrically rotate video, see :func:`simba.video_processors.egocentric_video_rotator.EgocentricVideoRotator`

    .. csv-table::
       :header: EXPECTED RUNTIMES
       :file: ../../../docs/tables/egocentrically_align_pose_cuda.csv
       :widths: 10, 45, 45
       :align: center
       :class: simba-table
       :header-rows: 1

    :param np.ndarray data: A 3D array of shape `(num_frames, num_points, 2)` containing 2D points for each frame. Each frame is represented as a 2D array of shape `(num_points, 2)`, where each row corresponds to a point's (x, y) coordinates.
    :param int anchor_1_idx: The index of the first anchor point in `data` used as the center of alignment. This body-part will be placed in the center of the image.
    :param int anchor_2_idx: The index of the second anchor point in `data` used to calculate the direction vector. This bosy-part will be located `direction` degrees from the anchor_1 body-part.
    :param int direction: The target direction in degrees to which the vector between the two anchors will be aligned.
    :param np.ndarray anchor_location: A 1D array of shape `(2,)` specifying the target (x, y) location for `anchor_1_idx` after alignment.
    :param int batch_size: Size of data that is processed on each iteration on GPU. default 1m. Increase if GPU allows.

    :return: A tuple containing the rotated data, and variables required for also rotating the video using the same rules:
             - `aligned_data`: A 3D array of shape `(num_frames, num_points, 2)` with the aligned 2D points.
             - `centers`: A 2D array of shape `(num_frames, 2)` containing the original locations of `anchor_1_idx` in each frame before alignment.
             - `rotation_vectors`: A 3D array of shape `(num_frames, 2, 2)` containing the rotation matrices applied to each frame.
    :rtype: Tuple[np.ndarray, np.ndarray, np.ndarray]

    :example:
    >>> DATA_PATH = r"/mnt/c/Users/sroni/OneDrive/Desktop/rotate_ex/data/501_MA142_Gi_Saline_0513.csv"
    >>> VIDEO_PATH = r"/mnt/c/Users/sroni/OneDrive/Desktop/rotate_ex/videos/501_MA142_Gi_Saline_0513.mp4"
    >>> SAVE_PATH = r"/mnt/c/Users/sroni/OneDrive/Desktop/rotate_ex/videos/501_MA142_Gi_Saline_0513_rotated.mp4"
    >>> ANCHOR_LOC = np.array([300, 300])
    >>>
    >>> df = read_df(file_path=DATA_PATH, file_type='csv')
    >>> bp_cols = [x for x in df.columns if not x.endswith('_p')]
    >>> data = df[bp_cols].values.reshape(len(df), int(len(bp_cols)/2), 2).astype(np.int64)
    >>> data, centers, rotation_matrices = egocentrically_align_pose_cuda(data=data, anchor_1_idx=6, anchor_2_idx=2, anchor_location=ANCHOR_LOC, direction=180,batch_size=36000000)
    """

    check_valid_array(data=data, source=egocentrically_align_pose_cuda.__name__, accepted_ndims=(3,), accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_int(name=f'{egocentrically_align_pose_cuda.__name__} anchor_1_idx', min_value=0, max_value=data.shape[1], value=anchor_1_idx)
    check_int(name=f'{egocentrically_align_pose_cuda.__name__} anchor_2_idx', min_value=0, max_value=data.shape[1], value=anchor_2_idx)
    if anchor_1_idx == anchor_2_idx: raise InvalidInputError(msg=f'Anchor 1 index ({anchor_1_idx}) cannot be the same as Anchor 2 index ({anchor_2_idx})', source=egocentrically_align_pose_cuda.__name__)
    check_int(name=f'{egocentrically_align_pose_cuda.__name__} direction', value=direction, min_value=0, max_value=360)
    check_valid_array(data=anchor_location, source=egocentrically_align_pose_cuda.__name__, accepted_ndims=(1,), accepted_axis_0_shape=[2,], accepted_dtypes=Formats.NUMERIC_DTYPES.value)
    check_int(name=f'{egocentrically_align_pose_cuda.__name__} batch_size', value=batch_size, min_value=1)
    results = np.full_like(a=data, fill_value=-1, dtype=np.int64)
    results_centers = np.full((data.shape[0], 2), fill_value=-1, dtype=np.int64)
    results_rotation_vectors = np.full((data.shape[0], 2, 2), fill_value=-1, dtype=np.float64)
    transposed_results_rotation_vectors = np.full((data.shape[0], 2, 2), fill_value=np.nan, dtype=np.float64)
    matrix_multiplier_arr = np.full((data.shape[0], data.shape[1], 2), fill_value=-1, dtype=np.int64)
    target_angle = np.deg2rad(direction)
    target_angle_dev = cuda.to_device(np.array([target_angle]))
    anchor_idx_dev = cuda.to_device(np.array([anchor_1_idx, anchor_2_idx]))
    anchor_loc_dev = cuda.to_device(anchor_location)

    for l in range(0, data.shape[0], batch_size):
        r = l + batch_size
        sample_data = np.ascontiguousarray(data[l:r]).astype(np.float64)
        sample_centers = np.ascontiguousarray(results_centers[l:r]).astype(np.int64)
        sample_rotation_vectors = np.ascontiguousarray(results_rotation_vectors[l:r].astype(np.float64))
        sample_transposed_rotation_vectors = np.ascontiguousarray(transposed_results_rotation_vectors[l:r])
        sample_matrix_multiplier_arr = np.ascontiguousarray(matrix_multiplier_arr[l:r])
        sample_results = np.ascontiguousarray(results[l:r].astype(np.float64))
        sample_data_dev = cuda.to_device(sample_data)
        sample_centers_dev = cuda.to_device(sample_centers)
        sample_matrix_multiplier_arr_dev = cuda.to_device(sample_matrix_multiplier_arr)
        sample_transposed_rotation_vectors_dev = cuda.to_device(sample_transposed_rotation_vectors)
        sample_rotation_vectors_dev = cuda.to_device(sample_rotation_vectors)
        sample_results_dev = cuda.to_device(sample_results)
        bpg = (sample_data.shape[0] + (THREADS_PER_BLOCK - 1)) // THREADS_PER_BLOCK
        _egocentric_align_kernel[bpg, THREADS_PER_BLOCK](sample_data_dev,
                                                         sample_centers_dev,
                                                         sample_rotation_vectors_dev,
                                                         sample_results_dev,
                                                         target_angle_dev,
                                                         anchor_idx_dev,
                                                         sample_transposed_rotation_vectors_dev,
                                                         sample_matrix_multiplier_arr_dev,
                                                         anchor_loc_dev)
        results[l:r] = sample_results_dev.copy_to_host()
        results_centers[l:r] = sample_centers_dev.copy_to_host()
        results_rotation_vectors[l:r] = sample_rotation_vectors_dev.copy_to_host()

    return results, results_centers, results_rotation_vectors


# DATA_PATH = r"/mnt/c/Users/sroni/OneDrive/Desktop/rotate_ex/data/501_MA142_Gi_Saline_0513.csv"
# VIDEO_PATH = r"/mnt/c/Users/sroni/OneDrive/Desktop/rotate_ex/videos/501_MA142_Gi_Saline_0513.mp4"
# SAVE_PATH = r"/mnt/c/Users/sroni/OneDrive/Desktop/rotate_ex/videos/501_MA142_Gi_Saline_0513_rotated.mp4"
# ANCHOR_LOC = np.array([300, 300])
# #
# # df = read_df(file_path=DATA_PATH, file_type='csv')
# # bp_cols = [x for x in df.columns if not x.endswith('_p')]
# # data = df[bp_cols].values.reshape(len(df), int(len(bp_cols)/2), 2).astype(np.int64)
# # data, centers, rotation_matrices = egocentrically_align_pose_cuda(data=data, anchor_1_idx=6, anchor_2_idx=2, anchor_location=ANCHOR_LOC, direction=180,batch_size=36000000)
# #
# for i in [250000, 500000, 1000000, 2000000, 4000000, 8000000, 16000000]:
#     data = np.random.randint(0, 500, (i, 6, 2))
#     times = []
#     for j in range(3):
#         start_t = time.perf_counter()
#         data, centers, rotation_matrices = egocentrically_align_pose_cuda(data=data, anchor_1_idx=6, anchor_2_idx=2, anchor_location=ANCHOR_LOC, direction=180, batch_size=36000000)
#         times.append(time.perf_counter() - start_t)
#     print(i, '\t' * 4, np.mean(times), '\t' * 4, np.std(times))


# from simba.video_processors.egocentric_video_rotator import EgocentricVideoRotator
#
# runner = EgocentricVideoRotator(video_path=VIDEO_PATH, centers=centers, rotation_vectors=rotation_matrices, anchor_location=(300, 300))
# runner.run()

#_, centers, rotation_vectors = egocentrically_align_pose(data=data, anchor_1_idx=6, anchor_2_idx=2, anchor_location=ANCHOR_LOC, direction=0)


__author__ = "Simon Nilsson"
__email__ = "sronilsson@gmail.com"

import os
from copy import copy, deepcopy
from typing import Dict, Optional, Union

import cv2
import numpy as np
from shapely.affinity import scale
from shapely.geometry import MultiPolygon, Polygon

try:
    from typing import Literal
except:
    from typing_extensions import Literal

import pandas as pd

from simba.mixins.geometry_mixin import GeometryMixin
from simba.mixins.image_mixin import ImageMixin
from simba.utils.checks import (check_float, check_if_dir_exists,
                                check_instance, check_int,
                                check_nvidea_gpu_available, check_str,
                                check_valid_boolean, is_img_bw)
from simba.utils.data import df_smoother, savgol_smoother
from simba.utils.enums import Formats, Methods, Options
from simba.utils.errors import FFMPEGCodecGPUError, InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (find_all_videos_in_directory, get_fn_ext,
                                    get_video_meta_data, remove_files,
                                    write_df)
from simba.video_processors.video_processing import (video_bg_subtraction,
                                                     video_bg_subtraction_mp)


class BlobLocationComputer(object):

    """
    Detecting and saving blob locations from video files.

    .. video:: _static/img/BlobLocationComputer.webm
       :width: 800
       :autoplay:
       :loop:

    :param Union[str, os.PathLike] data_path: Path to a video file or a directory containing video files. The videos will be processed for blob detection.
    :param Optional[bool] verbose:  If True, prints progress and success messages to the console. Default is True.
    :param Optional[bool] gpu: If True, GPU acceleration will be used for blob detection. Default is True.
    :param Optional[int] batch_size: The number of frames to process in each batch for blob detection. Default is 2500.
    :param Optional[Union[str, os.PathLike]] save_dir: Directory where the blob location data will be saved as CSV files. If None, the results will not be saved. Default is None.
    :param Optional[bool] multiprocessing: If True, video background subtraction will be done using  multiprocessing. Default is False.

    :example:
    >>> x = BlobLocationComputer(data_path=r"C:\troubleshooting\RAT_NOR\project_folder\videos\2022-06-20_NOB_DOT_4_downsampled_bg_subtracted.mp4", multiprocessing=True, gpu=True, batch_size=2000, save_dir=r"C:\blob_positions")
    >>> x.run()
    """
    def __init__(self,
                 data_path: Union[str, os.PathLike],
                 verbose: Optional[bool] = True,
                 gpu: Optional[bool] = True,
                 batch_size: int = 2500,
                 save_dir: Optional[Union[str, os.PathLike]] = None,
                 smoothing: Optional[str] = None,
                 multiprocessing: bool = False):


        if os.path.isdir(data_path):
            self.data_paths = find_all_videos_in_directory(directory=data_path, as_dict=True, raise_error=True).values()
        elif os.path.isfile(data_path):
            self.data_paths = [data_path]
        else:
            raise InvalidInputError(msg=f'{data_path} is not a valid directory or video file path or directory path.')
        self.video_meta_data = {}
        for i in self.data_paths:
            self.video_meta_data[get_fn_ext(filepath=i)[1]] = get_video_meta_data(video_path=i)
        if save_dir is not None:
            check_if_dir_exists(in_dir=os.path.dirname(save_dir), source=self.__class__.__name__)
        check_valid_boolean(value=[verbose, gpu, multiprocessing])
        if gpu and not check_nvidea_gpu_available():
            raise FFMPEGCodecGPUError(msg='No GPU detected.', source=self.__class__.__name__)
        if smoothing is not None:
            check_str(name=f'{self.__class__.__name__} smoothing', value=smoothing, options=Options.SMOOTHING_OPTIONS.value)
        check_int(name=f'{self.__class__.__name__} batch_size', value=batch_size, min_value=1)
        self.multiprocessing = multiprocessing
        self.verbose = verbose
        self.gpu = gpu
        self.batch_size = batch_size
        self.save_dir = save_dir
        self.smoothing = smoothing

    def run(self):
        timer = SimbaTimer(start=True)
        self.location_data = {}
        for file_cnt, video_path in enumerate(self.data_paths):
            video_timer = SimbaTimer(start=True)
            _, video_name, ext = get_fn_ext(filepath=video_path)
            temp_video_path = os.path.join(os.path.dirname(video_path), video_name + '_temp.mp4')
            if not self.multiprocessing:
                _ = video_bg_subtraction(video_path=video_path, verbose=self.verbose, bg_color=(0, 0, 0), fg_color=(255, 255, 255), save_path=temp_video_path)
            else:
                _ = video_bg_subtraction_mp(video_path=video_path, verbose=self.verbose, bg_color=(0, 0, 0), fg_color=(255, 255, 255), save_path=temp_video_path)
            self.location_data[video_name] = ImageMixin.get_blob_locations(video_path=temp_video_path, gpu=self.gpu, verbose=self.verbose, batch_size=self.batch_size).astype(np.int32)
            remove_files(file_paths=[temp_video_path])
            video_timer.stop_timer()
            print(f'Blob detection for video {video_name} ({file_cnt+1}/{len(self.data_paths)}) complete (elapsed time: {video_timer.elapsed_time_str}s)...')
        timer.stop_timer()
        if self.smoothing is not None:
            print('Smoothing data...')
            smoothened_data = {}
            if self.smoothing == Methods.SAVITZKY_GOLAY.value:
                for video_name, video_data in self.location_data.items():
                    smoothened_data[video_name] = savgol_smoother(data=video_data, fps=self.video_meta_data[video_name]['fps'], time_window=2000, source=video_name)
            if self.smoothing == Methods.GAUSSIAN.value:
                for video_name, video_data in self.location_data.items():
                    smoothened_data[video_name] = df_smoother(data=pd.DataFrame(video_data, columns=['X', 'Y']), fps=self.video_meta_data[video_name]['fps'], time_window=2000, source=video_name, method='gaussian')
            self.location_data = deepcopy(smoothened_data)
            del smoothened_data
        if self.save_dir is not None:
            for video_name, video_data in self.location_data.items():
                save_path = os.path.join(self.save_dir, f'{video_name}.csv')
                df = pd.DataFrame(video_data, columns=['X', 'Y'])
                write_df(df=df, file_type=Formats.CSV.value, save_path=save_path)
            if self.verbose:
                stdout_success(f'Video blob detection complete for {len(self.data_paths)} videos, data saved at {self.save_dir}', elapsed_time=timer.elapsed_time_str)
        else:
            if self.verbose:
                stdout_success(f'Video blob detection complete for {len(self.data_paths)} video', elapsed_time=timer.elapsed_time_str)





# x = BlobLocationComputer(data_path=r"C:\troubleshooting\RAT_NOR\project_folder\videos\2022-06-20_NOB_DOT_4_downsampled.mp4", multiprocessing=True, gpu=True, batch_size=2000, save_dir=r"C:\troubleshooting\RAT_NOR\project_folder\csv\blob_positions", smoothing='Savitzky Golay')
# x.run()



# x = BlobLocationComputer(data_path=r"C:\troubleshooting\RAT_NOR\project_folder\videos\2022-06-20_NOB_DOT_4_downsampled.mp4", multiprocessing=True, gpu=True, batch_size=2000, save_dir=r"C:\troubleshooting\RAT_NOR\project_folder\csv\blob_positions", smoothing='Savitzky Golay')
# x.run()


import os
from typing import Tuple, Union

import cv2
import numpy as np

from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_instance
from simba.utils.enums import TextOptions
from simba.utils.errors import InvalidInputError
from simba.utils.read_write import get_video_meta_data, read_frm_of_video

WIN_NAME = 'INTERACTIVE CLAHE - HIT ESC TO RUN'
CLIP_LIMIT = 'CLIP LIMIT'
TILE_SIZE = 'TILE SIZE'
SELECT_VIDEO_FRAME = 'SHOW FRAME'

def interactive_clahe_ui(data: Union[str, os.PathLike]) -> Tuple[float, int]:
    """
    Create a user interface using OpenCV to explore and set appropriate CLAHE settings tile size and clip limit.

    .. image:: _static/img/interactive_clahe_ui.gif
       :width: 500
       :align: center

    :param Union[str, os.PathLike, np.ndarray] data: Path to a video file.
    :return Tuple[float, int]: Tuple containing the chosen clip limit and tile size.

    :example:
    >>> video = cv2.imread(r"D:\EPM\sample_2\video_1.mp4")
    >>> interactive_clahe_ui(data=video)
    """
    global original_img, font_size, x_spacer, y_spacer, txt

    def _get_trackbar_values(v):
        global original_img, font_size, x_spacer, y_spacer, txt
        clip_limit = cv2.getTrackbarPos(CLIP_LIMIT, WIN_NAME) / 10.0
        tile_size = cv2.getTrackbarPos(TILE_SIZE, WIN_NAME)
        if tile_size % 2 == 0: tile_size += 1
        clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=(tile_size, tile_size))
        img_clahe = clahe.apply(original_img)
        cv2.putText(img_clahe, txt,(TextOptions.BORDER_BUFFER_X.value, TextOptions.BORDER_BUFFER_Y.value + y_spacer), TextOptions.FONT.value, font_size, (255, 255, 255), 3)
        cv2.imshow(WIN_NAME, img_clahe)
        cv2.waitKey(100)

    def _change_img(v):
        global original_img, font_size, x_spacer, y_spacer, txt
        new_frm_id = cv2.getTrackbarPos(SELECT_VIDEO_FRAME, WIN_NAME)
        original_img = read_frm_of_video(video_path=data, frame_index=new_frm_id, greyscale=True)
        clip_limit = cv2.getTrackbarPos(CLIP_LIMIT, WIN_NAME) / 10.0
        tile_size = cv2.getTrackbarPos(TILE_SIZE, WIN_NAME)
        if tile_size % 2 == 0: tile_size += 1
        clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=(tile_size, tile_size))
        img_clahe = clahe.apply(original_img)
        cv2.putText(img_clahe, txt,(TextOptions.BORDER_BUFFER_X.value, TextOptions.BORDER_BUFFER_Y.value + y_spacer), TextOptions.FONT.value, font_size, (255, 255, 255), 3)
        cv2.imshow(WIN_NAME, img_clahe)
        cv2.waitKey(100)

    check_instance(source=interactive_clahe_ui.__name__, instance=data, accepted_types=(np.ndarray, str))
    if isinstance(data, str):
        video_meta_data = get_video_meta_data(video_path=data)
        original_img = read_frm_of_video(video_path=data, frame_index=0, greyscale=True)
    else:
        raise InvalidInputError(msg=f'data has to be a path to a video file, but got {type(data)}', source=interactive_clahe_ui.__name__)
    txt = 'Hit ESC to run with chosen settings'
    font_size, x_spacer, y_spacer = PlottingMixin().get_optimal_font_scales(text=txt, accepted_px_width=int(video_meta_data["width"] / 2), accepted_px_height=int(video_meta_data["height"] / 15), text_thickness=3)

    img = np.copy(original_img)
    cv2.putText(img, txt, (TextOptions.BORDER_BUFFER_X.value, TextOptions.BORDER_BUFFER_Y.value + y_spacer), TextOptions.FONT.value, font_size, (255, 255, 255), 3)
    cv2.namedWindow(WIN_NAME, cv2.WINDOW_NORMAL)
    cv2.resizeWindow(WIN_NAME, video_meta_data['width'], video_meta_data['height'])
    cv2.imshow(WIN_NAME, img)
    cv2.createTrackbar(CLIP_LIMIT, WIN_NAME, 10, 300, _get_trackbar_values)
    cv2.createTrackbar(TILE_SIZE, WIN_NAME,  8, 64,  _get_trackbar_values)
    cv2.createTrackbar(SELECT_VIDEO_FRAME, WIN_NAME, 0, video_meta_data['frame_count'], _change_img)

    while True:
        if cv2.getWindowProperty(WIN_NAME, cv2.WND_PROP_VISIBLE) < 1:
            cv2.destroyAllWindows()
            break
        k = cv2.waitKey(1) & 0xFF
        if k == 27:
            clip_limit = cv2.getTrackbarPos(CLIP_LIMIT, WIN_NAME) / 10.0
            tile_size = cv2.getTrackbarPos(TILE_SIZE, WIN_NAME)
            if tile_size % 2 == 0: tile_size += 1
            cv2.destroyAllWindows()
            return clip_limit, tile_size


#interactive_clahe_ui(data=r"D:\EPM\sample_2\video_1.mp4")

# # Function to update CLAHE
# def update_clahe(x):
#     global img, clahe
#     clip_limit = cv2.getTrackbarPos('Clip Limit', 'CLAHE') / 10.0  # Scale the trackbar value
#     tile_size = cv2.getTrackbarPos('Tile Size', 'CLAHE')
#     if tile_size % 2 == 0:
#         tile_size += 1  # Ensure tile size is odd
#     clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=(tile_size, tile_size))
#     img_clahe = clahe.apply(img)
#     cv2.imshow('CLAHE', img_clahe)
#
# # Load an image
# img = cv2.imread('/Users/simon/Downloads/PXL_20240429_222923838.jpg', cv2.IMREAD_GRAYSCALE)
#
# # Create a window
# cv2.namedWindow('CLAHE', cv2.WINDOW_NORMAL)
#
# # Initialize the clip limit trackbar
# cv2.createTrackbar('Clip Limit', 'CLAHE', 10, 300, update_clahe)
#
# # Initialize the tile size trackbar
# cv2.createTrackbar('Tile Size', 'CLAHE', 8, 64, update_clahe)
#
# # Apply CLAHE with initial parameters
# clahe = cv2.createCLAHE(clipLimit=1.0, tileGridSize=(8, 8))
# img_clahe = clahe.apply(img)
# cv2.imshow('Original', img)
# cv2.imshow('CLAHE', img_clahe)
#
# cv2.waitKey(0)
# cv2.destroyAllWindows()



import glob
import os
from tkinter import *
from typing import Union

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.mixins.statistics_mixin import Statistics
from simba.mixins.unsupervised_mixin import UnsupervisedMixin
from simba.ui.tkinter_functions import FolderSelect
from simba.unsupervised.cluster_validation_calculator import ClusterValidators
from simba.unsupervised.dbcv_calculator import DBCVCalculator
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_dir_exists,
                                check_if_filepath_list_is_empty)
from simba.utils.enums import Formats
from simba.utils.errors import CountError, InvalidInputError
from simba.utils.printing import SimbaTimer, stdout_success

VALIDATION_METRICS = [
    "DENSITY-BASED CLUSTER VALIDATION (DBCV)",
    "DUNN INDEX",
    "DAVIS-BUILDIN INDEX",
    "CALINSKI-HARABASZ SCORE",
]
DBCV = "DENSITY-BASED CLUSTER VALIDATION (DBCV)"
DUNN = "DUNN INDEX"
DAVIS_BUILDIN = "DAVIS-BUILDIN INDEX"
CALINSKI_HARABASZ = "CALINSKI-HARABASZ SCORE"


class ClusterValidatorPopUp(PopUpMixin, ConfigReader, UnsupervisedMixin):
    def __init__(self, config_path: Union[str, os.PathLike]):
        check_file_exist_and_readable(file_path=config_path)
        PopUpMixin.__init__(self, title="CLUSTER VALIDATION METRICS")
        ConfigReader.__init__(self, config_path=config_path)
        UnsupervisedMixin.__init__(self)
        self.data_frm = LabelFrame(
            self.main_frm,
            text="DATA",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.folder_selected = FolderSelect(
            self.data_frm,
            "DATASETS (DIRECTORY WITH PICKLES):",
            lblwidth=35,
            initialdir=self.project_path,
        )
        self.data_frm.grid(row=0, column=0, sticky=NW)
        self.folder_selected.grid(row=0, column=0, sticky=NW)
        self.cb_settings = self.create_cb_frame(
            main_frm=self.main_frm,
            cb_titles=VALIDATION_METRICS,
            frm_title="VALIDATION METHODS",
        )
        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        timer = SimbaTimer(start=True)
        check_if_dir_exists(in_dir=self.folder_selected.folder_path)
        data_paths = glob.glob(
            self.folder_selected.folder_path + f"/*.{Formats.PICKLE.value}"
        )
        check_if_filepath_list_is_empty(
            filepaths=data_paths,
            error_msg=f"No pickle files found in {self.folder_selected.folder_path}",
        )
        selections = []
        for k, v in self.cb_settings.items():
            if v.get():
                selections.append(k)
        if len(selections) == 0:
            raise CountError(
                msg="Select at least 1 data-type checkbox",
                source=self.__class__.__name__,
            )
        for selection in selections:
            if selection == DBCV:
                print(f"Running {DBCV}...")
                dbcv_calculator = DBCVCalculator(
                    data_path=self.folder_selected.folder_path,
                    config_path=self.config_path,
                )
                dbcv_calculator.run()
            else:
                if selection == DUNN:
                    validator_func = Statistics.dunn_index
                elif selection == DAVIS_BUILDIN:
                    validator_func = Statistics.davis_bouldin
                elif selection == CALINSKI_HARABASZ:
                    validator_func = Statistics.calinski_harabasz
                else:
                    raise InvalidInputError(
                        msg=f"Validator algorithm {selection} is not a valid options. Options: {VALIDATION_METRICS}",
                        source=self.__class__.__name__,
                    )
                print(f"Running {selection}...")
                calculator = ClusterValidators(
                    config_path=self.config_path,
                    data_path=self.folder_selected.folder_path,
                    validator_func=validator_func,
                )
                calculator.run()
                calculator.save()
        timer.stop_timer()
        stdout_success(
            msg="Cluster validations complete!", source=timer.elapsed_time_str
        )


# _ = ClusterValidatorPopUp(config_path='/Users/simon/Desktop/envs/NG_Unsupervised/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

from tkinter import *

import numpy as np

from simba.mixins.pop_up_mixin import PopUpMixin
from simba.third_party_label_appenders.transform.sleap_csv_to_yolo import \
    Sleap2Yolo
from simba.third_party_label_appenders.transform.sleap_to_yolo import \
    SleapAnnotations2Yolo
from simba.third_party_label_appenders.transform.utils import \
    get_yolo_keypoint_flip_idx
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, FolderSelect,
                                        SimbaCheckbox, SimBADropDown)
from simba.utils.checks import check_if_dir_exists
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    str_2_bool)

TRAIN_SIZE_OPTIONS = list(np.arange(10, 110, 10))
SAMPLE_SIZE_OPTIONS = list(np.arange(50, 650, 50))

PADDING_OPTIONS = list(np.round(np.arange(0.01, 10.05, 0.05),2).astype(str))
PADDING_OPTIONS = list(np.insert(PADDING_OPTIONS, 0, 'None'))



class SLEAPAnnotations2YoloPopUp(PopUpMixin):

    """
    GUI for converting DeepLabCut (DLC) annotations to YOLOv8 keypoint format.

    :example:
    >>> SLEAPAnnotations2YoloPopUp()
    """

    def __init__(self):
        PopUpMixin.__init__(self, title="SLEAP ANNOTATIONS TO YOLO POSE ESTIMATION ANNOTATIONS", icon='sleap_small')
        settings_frm = CreateLabelFrameWithIcon(parent=self.main_frm, header="SETTINGS", icon_name='settings')
        self.sleap_dir = FolderSelect(settings_frm, folderDescription="SLEAP DATA DIRECTORY (.SLP):", lblwidth=35, entry_width=40, initialdir=r"D:\troubleshooting\two_animals_sleap\import_data")
        self.save_dir = FolderSelect(settings_frm, folderDescription="SAVE DIRECTORY:", lblwidth=35, entry_width=40, initialdir=r"D:\troubleshooting\two_animals_sleap\yolo_kpts_2")

        self.verbose_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="VERBOSE: ", label_width=35, dropdown_width=40, value='TRUE')
        self.train_size_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=TRAIN_SIZE_OPTIONS, label="TRAIN SIZE (%): ", label_width=35, dropdown_width=40, value=70)
        self.grey_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="GREYSCALE: ", label_width=35, dropdown_width=40, value='FALSE')
        self.padding_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=PADDING_OPTIONS, label="PADDING: ", label_width=35, dropdown_width=40, value='None')
        self.clahe_dropdown = SimBADropDown(parent=settings_frm, dropdown_options=['TRUE', 'FALSE'], label="CLAHE: ", label_width=35, dropdown_width=40, value='FALSE')
        single_id_cb, self.single_id_var = SimbaCheckbox(parent=settings_frm, txt="REMOVE ANIMAL ID'S", val=False)


        settings_frm.grid(row=0, column=0, sticky=NW)
        self.sleap_dir .grid(row=0, column=0, sticky=NW)
        self.save_dir.grid(row=1, column=0, sticky=NW)

        self.verbose_dropdown.grid(row=2, column=0, sticky=NW)
        self.train_size_dropdown.grid(row=3, column=0, sticky=NW)
        self.grey_dropdown.grid(row=4, column=0, sticky=NW)
        self.clahe_dropdown.grid(row=5, column=0, sticky=NW)
        self.padding_dropdown.grid(row=6, column=0, sticky=NW)
        single_id_cb.grid(row=7, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()


    def run(self):
        sleap_dir = self.sleap_dir.folder_path
        save_dir = self.save_dir.folder_path

        check_if_dir_exists(in_dir=sleap_dir, source=f'{self.__class__.__name__} SLEAP DATA DIRECTORY', raise_error=True)
        check_if_dir_exists(in_dir=save_dir, source=f'{self.__class__.__name__} SAVE DIRECTORY', raise_error=True)
        _ = find_files_of_filetypes_in_directory(directory=sleap_dir, extensions=['.slp'], raise_error=True)

        grey = str_2_bool(self.grey_dropdown.get_value())
        train_size = int(self.train_size_dropdown.get_value()) / 100
        verbose = str_2_bool(self.verbose_dropdown.get_value())
        clahe = str_2_bool(self.clahe_dropdown.get_value())
        padding = float(self.padding_dropdown.get_value()) if self.padding_dropdown.get_value() != 'None' else 0.0
        single_id = 'animal_1' if self.single_id_var.get() else None

        runner = SleapAnnotations2Yolo(sleap_dir=sleap_dir, save_dir=save_dir, padding=padding, train_size=train_size, verbose=verbose, greyscale=grey, clahe=clahe, single_id=single_id)
        runner.run()


#SLEAPAnnotations2YoloPopUp()

